"""The joblib htcondor UI I/O."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import json
from copy import copy
from datetime import datetime
from json.decoder import JSONDecodeError
from pathlib import Path

from ..backend import (
    _BackendMeta,
    _TaskStatus,
)
from .uilogging import logger


class MetaTree:
    """Class for managing a metadata tree.

    Parameters
    ----------
    meta : _BackendMeta
        The backend metadata object to use as a root.
    fname : pathlib.Path
        The path to the JSON file

    """

    def __init__(self, meta: _BackendMeta, fname) -> None:
        self.meta: _BackendMeta = meta
        self.children: list[MetaTree] = []
        if not isinstance(fname, Path):
            fname = Path(fname)
        self.fname = fname
        self._task_status = {
            "done": 0,
            "running": 0,
            "sent": 0,
            "queued": 0,
            "total": 0,
        }
        self._core_hours = 0
        self._level_status_summary = [
            {
                "done": 0,
                "running": 0,
                "sent": 0,
                "queued": 0,
                "total": 0,
                "throttle": -1,
            }
        ]
        self._size = 0
        self._depth = 0
        self._last_update = datetime.now()

    @classmethod
    def from_json(cls: type["MetaTree"], fname: Path) -> "MetaTree":
        """Load object from JSON.

        Parameters
        ----------
        cls : MetaTree type
            The type of instance to create.
        fname : pathlib.Path
            The path to the JSON file.

        Returns
        -------
        MetaTree instance
            The initialised object instance.

        """
        with open(fname) as f:
            meta = _BackendMeta.from_json(json.load(f))
        return cls(meta, fname)

    def _update_from_list(self, all_meta: list["MetaTree"]) -> None:
        """Update from list of MetaTrees nodes.

        Parameters
        ----------
        all_meta : list of MetaTree
            List of nodes to update from.

        Raises
        ------
        OSError
            If file could not be opened.

        """
        # Reload the json file
        child_uuids = [c.meta.uuid for c in self.children]
        for tree in all_meta:
            if (
                tree.meta.parent is not None
                and tree.meta.parent == self.meta.uuid
                and tree.meta.uuid not in child_uuids
            ):
                logger.log(
                    level=9,
                    msg=f"Appending {tree.meta.uuid} to {self.meta.uuid}",
                )
                self.children.append(tree)
        for c in self.children:
            c._update_from_list(all_meta)
        # Update the task status
        self._update_task_status()
        self._update_level_status_summary()
        self._update_core_hours()
        self._update_metrics()

    def _update_metrics(self) -> None:
        """Update the tree metrics."""
        self._size = 1
        last_current = self.meta.update_timestamp
        max_depth = 0
        for c in self.children:
            last_current = max(last_current, c.last_update())
            self._size += c.size()
            max_depth = max(max_depth, c.depth())
        self._last_update = last_current
        self._depth = 1 + max_depth

    def update(self) -> None:
        """Update the tree.

        Raises
        ------
        OSError
            If the JSON file could not be opened.

        """
        # Read all the json files in the directory
        logger.log(level=10, msg=f"Updating tree {self.meta.uuid}")
        try:
            with self.fname.open("r") as fd:
                self.meta = _BackendMeta.from_json(json.load(fd))
        except OSError as e:
            logger.error(f"Error loading {self.fname}: {e}")
            return
        logger.log(
            level=10, msg=f"Updating list of files for {self.meta.uuid}"
        )
        all_meta = []
        for f in self.fname.parent.glob("*.json"):
            try:
                all_meta.append(MetaTree.from_json(f))
            except OSError as e:
                logger.error(f"Error loading {f}: {e}")
                continue
            except JSONDecodeError as e:
                logger.error(f"Error decoding {f}: {e}")
                continue
        all_meta = sorted(all_meta, key=lambda x: x.meta.start_timestamp)
        # Update the tree from this list of files
        logger.log(
            level=10, msg=f"Updating children from files {self.meta.uuid}"
        )
        self._update_from_list(all_meta)
        logger.log(level=10, msg=f"Tree updated {self.meta.uuid}")

    def __repr__(self) -> str:
        """Representation of object."""
        out = f"MetaTree({self.meta.uuid})"
        if len(self.children) > 0:
            out += f" [{len(self.children)}]-> \n"
            for c in self.children:
                out += f"\t{c}\n"
        return out

    def size(self) -> int:
        """Size of the tree.

        Returns
        -------
        int
            The size of the tree.

        """
        return self._size

    def depth(self) -> int:
        """Depth of the tree.

        Returns
        -------
        int
            The depth of the tree.

        """
        return self._depth

    def last_update(self) -> datetime:
        """Get last update time.

        Returns
        -------
        datetime
            The time of the last update.

        """
        return self._last_update

    def get_level_status_summary(self) -> list[dict[str, int]]:
        """Get status summary of current level.

        Returns
        -------
        list
            The overall status summary of the current level as a list of dicts
            with all the status counters.

        """
        return self._level_status_summary

    def _update_level_status_summary(self) -> None:
        """Update level status summary and store in private variable."""
        this_level_summary = [copy(self._task_status)]
        this_level_summary[0]["throttle"] = self.meta.throttle
        child_level_summary = [
            x.get_level_status_summary() for x in self.children
        ]
        if len(child_level_summary) > 0:
            n_child_levels = max([len(x) for x in child_level_summary])
            for i in range(n_child_levels):
                this_level_summary.append(
                    {
                        "done": 0,
                        "running": 0,
                        "sent": 0,
                        "queued": 0,
                        "total": 0,
                        "throttle": 0,
                    }
                )
                for t_child_summary in child_level_summary:
                    if i < len(t_child_summary):
                        for k, v in t_child_summary[i].items():
                            if k != "throttle":
                                this_level_summary[-1][k] += v
                            else:
                                this_level_summary[-1][k] = v
        self._level_status_summary = this_level_summary

    def get_task_status(self) -> dict[str, int]:
        """Get task status.

        Returns
        -------
        dict
            The overall status of the jobs as a dict.

        """
        return self._task_status

    def _update_task_status(self) -> None:
        """Update task status and store in private variable."""
        task_status = [x.get_status() for x in self.meta.task_status]
        n_queued = task_status.count(_TaskStatus.QUEUED)
        n_sent = task_status.count(_TaskStatus.SENT)
        n_running = task_status.count(_TaskStatus.RUN)
        n_done = task_status.count(_TaskStatus.DONE)
        self._task_status = {
            "done": n_done,
            "running": n_running,
            "sent": n_sent,
            "queued": n_queued,
            "total": self.meta.n_tasks,
        }
        # logger.debug(f"Task status: {out}")

    def get_core_hours(self) -> float:
        """Get total core hours.

        Returns
        -------
        float
            Total core hours.

        """
        return self._core_hours

    def _update_core_hours(self) -> None:
        """Update core hours and store in private variable."""
        core_hours = 0
        for task in self.meta.task_status:
            if task.done_timestamp is not None:
                delta = task.done_timestamp - (
                    task.run_timestamp
                    if task.run_timestamp is not None
                    else task.sent_timestamp
                )  # type: ignore
                core_hours += delta.total_seconds() / 3600 * task.request_cpus
        for c in self.children:
            core_hours += c.get_core_hours()
        self._core_hours = core_hours


def parse(root_fname) -> MetaTree:
    """Parse meta tree from file.

    Parameters
    ----------
    root_fname : pathlib.Path
        The path to the root JSON file.

    Returns
    -------
    MetaTree instance
        The parsed tree

    """
    logger.log(level=10, msg=f"Parsing {root_fname}")
    tree = MetaTree.from_json(root_fname)
    tree.update()
    return tree



"""The joblib htcondor UI base Window class."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

from typing import Any


class Window:
    """Base class for a UI window.

    Parameters
    ----------
    y : int
        The y coordinate of the window.
    x : int
        The x coordinate of the window.
    window : curses.window
        The curses window object.

    """

    def __init__(self, y: int, x: int, window: Any) -> None:
        self.y = y
        self.x = x
        self.h, self.w = window.getmaxyx()
        self.win = window

    def render(self) -> None:
        """Render the window."""
        pass

    def refresh(self) -> None:
        """Refresh the window."""
        self.win.refresh()

    def scroll_up(self) -> None:
        """Handle scroll up."""
        pass

    def scroll_down(self) -> None:
        """Handle scroll down."""
        pass

    def scroll_left(self) -> None:
        """Handle scroll left."""
        pass

    def scroll_right(self) -> None:
        """Handle scroll right."""
        pass

    def action(self, action) -> None:
        """Trigger action.

        Parameters
        ----------
        action : str
            The action to trigger.

        """
        pass

    def resize(self, w=None, h=None) -> None:
        """Resize window.

        Parameters
        ----------
        w : int
            The new width. If None, the maximum width is used.
        h : int
            The new height. If None, the maximum height is used.

        """
        maxy, maxx = self.win.getmaxyx()
        if w is None:
            w = maxx - self.y
        if h is None:
            h = maxy - self.x
        self.h = h
        self.w = w
        self.win.resize(self.h, self.w)

    def get_menu(self) -> list[tuple[int, str]]:
        """Get context menu.

        Returns
        -------
        list
            The context menu. Each item is a tuple with the first element
            being the index of the letter that triggers the action and the
            second the name of the action.

        """
        return []



"""Provide tests for _HTCondorBackend."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import logging

from joblib.externals.cloudpickle import cloudpickle as pickle

from joblib_htcondor.backend import _HTCondorBackend


def compare_backends(a, b) -> bool:
    """Compare two backends."""
    ok = True
    ok = ok and a.__class__ == b.__class__
    core_vars = [
        "_pool",
        "_schedd",
        "_universe",
        "_python_path",
        "_request_cpus",
        "_request_memory",
        "_request_disk",
        "_initial_dir",
        "_log_dir_prefix",
        "_poll_interval",
        "_shared_data_dir",
        "_extra_directives",
        "_worker_log_level",
        "_throttle",
    ]
    for var in core_vars:
        ok = ok and getattr(a, var) == getattr(b, var)
    return ok


def test_pickle() -> None:
    """Test pickling of the backend."""
    backend = _HTCondorBackend(request_cpus=1, request_memory="2GB")
    pickled_backend = pickle.loads(pickle.dumps(backend))
    assert compare_backends(pickled_backend, backend)

    backend2 = _HTCondorBackend(
        pool="head2.htc.inm7.de",
        request_cpus=1,
        request_memory="8GB",
        request_disk="1GB",
        python_path="/home/fraimondo/miniconda3/ppc64le_dev/bin/python",
        extra_directives={"Requirements": 'Arch == "ppc64le"'},
        worker_log_level=logging.DEBUG,
        throttle=11,
        poll_interval=0.1,
    )
    pickled_backend2 = pickle.loads(pickle.dumps(backend2))
    assert compare_backends(pickled_backend2, backend2)



"""The joblib htcondor backend implementation."""

# Authors: Synchon Mandal <s.mandal@fz-juelich.de>
#          Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import json
import logging
import shutil
import signal
import sys
import time
import traceback
from collections import deque
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass, field
from datetime import datetime
from enum import IntEnum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Union,
)
from uuid import uuid1

import htcondor2
from joblib.parallel import (
    ParallelBackendBase,
    SequentialBackend,
    register_parallel_backend,
)

from .delayed_submission import DelayedSubmission
from .logging import logger


if TYPE_CHECKING:
    from multiprocessing.pool import AsyncResult

    from classad2 import ClassAd
    from joblib import Parallel
    from joblib.parallel import BatchedCalls


__all__ = ["register"]


def register() -> None:
    """Register joblib htcondor backend."""
    register_parallel_backend("htcondor", _HTCondorBackend)


class _TaskStatus(IntEnum):
    """IntEnum for task status."""

    QUEUED = 0
    SENT = 1
    RUN = 2
    DONE = 3


@dataclass
class _TaskMeta:
    """Joblib task metadata."""

    cluster_id: Optional[int] = field(default=None)
    queued_timestamp: datetime = field(default_factory=datetime.now)
    sent_timestamp: Optional[datetime] = field(default=None)
    run_timestamp: Optional[datetime] = field(default=None)
    done_timestamp: Optional[datetime] = field(default=None)
    request_cpus: int = field(default=0)

    def asdict(self) -> dict[str, Any]:
        """Represent object as dictionary.

        Returns
        -------
        dict
            The object as a dictionary.

        """
        return {
            "cluster_id": self.cluster_id,
            "queued_timestamp": self.queued_timestamp.isoformat(),
            "sent_timestamp": self.sent_timestamp.isoformat()
            if self.sent_timestamp is not None
            else None,
            "run_timestamp": self.run_timestamp.isoformat()
            if self.run_timestamp is not None
            else None,
            "done_timestamp": self.done_timestamp.isoformat()
            if self.done_timestamp is not None
            else None,
            "request_cpus": self.request_cpus,
        }

    def get_status(self) -> int:
        """Return the status of the task.

        Returns
        -------
        int
            The status of the task.

        """
        if self.done_timestamp is not None:
            return _TaskStatus.DONE
        if self.run_timestamp is not None:
            return _TaskStatus.RUN
        if self.sent_timestamp is not None:
            return _TaskStatus.SENT
        return _TaskStatus.QUEUED

    def update_run_from_file(self, run_fname: Path) -> bool:
        """Update object from run file.

        Parameters
        ----------
        run_fname : pathlib.Path
            The run file to update object with.

        Returns
        -------
        bool
            Whether the object was updated.

        Raises
        ------
        OSError
            If data could not be updated from run file.

        """
        if not run_fname.exists():
            return False
        out = False
        if self.run_timestamp is None:
            try:
                with run_fname.open("r") as f:
                    run_ts = datetime.fromisoformat(f.readline())
                    self.run_timestamp = run_ts
                    out = True
            except OSError as e:
                logger.warning(
                    f"Error reading run timestamp from {run_fname}: {e}"
                )
            except ValueError as e:
                logger.warning(
                    f"Error parsing run timestamp from {run_fname}: {e}"
                )
        return out

    @classmethod
    def from_json(cls: type["_TaskMeta"], data: dict[str, Any]) -> "_TaskMeta":
        """Load object from JSON.

        Parameters
        ----------
        cls : _TaskMeta instance
            The type of instance to create.
        data : dict
            The data to initialise object with.

        Returns
        -------
        _TaskMeta instance
            The initialised object instance.

        """
        return cls(
            cluster_id=data["cluster_id"],
            queued_timestamp=datetime.fromisoformat(data["queued_timestamp"]),
            sent_timestamp=datetime.fromisoformat(data["sent_timestamp"])
            if data["sent_timestamp"] is not None
            else None,
            run_timestamp=datetime.fromisoformat(data["run_timestamp"])
            if data["run_timestamp"] is not None
            else None,
            done_timestamp=datetime.fromisoformat(data["done_timestamp"])
            if data["done_timestamp"] is not None
            else None,
            request_cpus=data.get("request_cpus", 1),
        )


@dataclass
class _BackendMeta:
    """Class for metadata I/O.

    This class defines the interface for writing and reading metadata files
    stored as JSON in the filesystem. Files are written by the backend and
    used by the UI machinery.

    """

    uuid: str
    parent: Optional[str] = field(default=None)  # Parent backend batch id
    recursion_level: int = field(default=0)  # Recursion level of the backend
    throttle: int = 0  # Throttle value for the backend
    shared_data_dir: Optional[Path] = field(
        default=None
    )  # Shared data directory for the backend
    n_tasks: int = field(default=0)  # Number of tasks received
    start_timestamp: datetime = field(
        default_factory=datetime.now
    )  # Start timestamp of the backend
    update_timestamp: datetime = field(
        default_factory=datetime.now
    )  # Update timestamp of the backend

    # Status of the tasks
    task_status: list[_TaskMeta] = field(default_factory=list)

    def asdict(self) -> dict[str, Any]:
        """Represent object as dictionary.

        Returns
        -------
        dict
            The object as a dictionary.

        """
        return {
            "uuid": self.uuid,
            "parent": self.parent,
            "recursion_level": self.recursion_level,
            "throttle": self.throttle,
            "shared_data_dir": self.shared_data_dir.as_posix()
            if self.shared_data_dir is not None
            else None,
            "n_tasks": self.n_tasks,
            "start_timestamp": self.start_timestamp.isoformat(),
            "update_timestamp": self.update_timestamp.isoformat(),
            "task_status": [ts.asdict() for ts in self.task_status],
        }

    @classmethod
    def from_json(
        cls: type["_BackendMeta"], data: dict[str, Any]
    ) -> "_BackendMeta":
        """Load object from JSON.

        Parameters
        ----------
        cls : _BackendMeta instance
            The type of instance to create.
        data : dict
            The data to initialise object with.

        Returns
        -------
        _BackendMeta instance
            The initialised object instance.

        """
        obj = cls(
            uuid=data["uuid"],
            parent=data["parent"],
            recursion_level=data["recursion_level"],
            throttle=data["throttle"],
            shared_data_dir=Path(data["shared_data_dir"]),
            n_tasks=data["n_tasks"],
            start_timestamp=datetime.fromisoformat(data["start_timestamp"]),
            update_timestamp=datetime.fromisoformat(data["update_timestamp"]),
        )
        obj.task_status = [
            _TaskMeta.from_json(ts) for ts in data["task_status"]
        ]
        return obj


@dataclass
class _HTCondorJobMeta:
    """Class for keeping track of HTCondor job submissions."""

    tracking_future: Future
    htcondor_submit: htcondor2.Submit
    htcondor_submit_result: Optional[htcondor2.SubmitResult]
    callback: Optional[Callable]
    pickle_fname: Path
    delayed_submission: DelayedSubmission
    task_id: int = 0  # Internal task ID
    cluster_id = None  # HTCondor cluster ID

    def __repr__(self) -> str:
        submit_info = ""
        if self.htcondor_submit_result:
            submit_info = (
                f"[ClusterId: {self.htcondor_submit_result.cluster()}]"
            )

        out = (
            f"<HTCondorJobMeta(pickle_fname={self.pickle_fname}){submit_info}>"
        )
        return out


class _HTCondorBackend(ParallelBackendBase):
    """Class for HTCondor backend for joblib.

    Parameters
    ----------
    request_cpus : int
        HTCondor CPUs to request.
    request_memory : str
        HTCondor memory to request.
    pool : str, classad2.ClassAd, list of str or None, optional
        Pool to initiate htcondor2.Collector client with (default None).
    schedd : htcondor2.Schedd or None, optional
        Scheduler to use for submitting jobs (default None).
    universe : str, optional
        HTCondor universe to use (default "vanilla").
    python_path : str, optional
        Path to the Python binary (defaults to current python executable).
    request_disk : str, optional
        HTCondor disk to request (default "8G").
    initial_dir : str or pathlib.Path or None, optional
        HTCondor initial directory for job. If None, will resolve to current
        working directory (default None).
    log_dir_prefix : str or None, optional
        Prefix for the log directory. The directory prefix needs
        to exist before submitting as HTCondor demands. If None, will resolve
        to `"`initial_dir`/logs/<generated unique batch name>"`
        (default None).
    poll_interval : int, optional
        Interval in seconds to poll the scheduler for job status
        (default 5).
    shared_data_dir : str or pathlib.Path or None, optional
        Directory to store shared data between jobs. If None, will resolve to
        `"<current working directory>/joblib_htcondor_shared_data"`
        (default None).
    extra_directives : dict or None, optional
        Extra directives to pass to the HTCondor submit file (default None).
    worker_log_level : int, optional
        Log level for the worker (default is logger.INFO).
    throttle : int or list of int or None, optional
        Throttle the number of jobs submitted at once. If list, the first
        element is the throttle for the current level and the rest are
        for the nested levels (default None).
    batch_size : int, optional
        Number of joblib jobs to group together in a single HTCondor job (
        default 1).
    max_recursion_level : int, optional
        Maximum recursion level of the backend. Once the
        recursion level reaches this value, the backend will switch to a
        Sequential backend. If -1, the switching is disabled. If 0,
        no recursion is allowed (default 0).
    export_metadata : bool, optional
        Export metadata to a file, to be used with the HTCondor Joblib Monitor.
        This increases the load on the filesystem considerably if the number
        of jobs is high and the duration is short (default False).

    Raises
    ------
    ValueError
        If `throttle` is an empty list.
    RuntimeError
        If htcondor2.Schedd client cannot be created.

    """

    def __init__(  # noqa: C901
        self,
        request_cpus: int,
        request_memory: str,
        pool: Union[str, "ClassAd", list[str], None] = None,
        schedd: Optional[htcondor2.Schedd] = None,
        universe: str = "vanilla",
        python_path: Optional[str] = None,
        request_disk: str = "0GB",
        initial_dir: Union[str, Path, None] = None,
        log_dir_prefix: Optional[str] = None,
        poll_interval: int = 5,
        shared_data_dir: Union[str, Path, None] = None,
        extra_directives: Optional[dict] = None,
        worker_log_level: int = logging.INFO,
        throttle: Union[int, list[int], None] = None,
        batch_size: int = 1,
        max_recursion_level: int = 0,
        export_metadata: bool = False,
    ) -> None:
        super().__init__()

        logger.debug("Initializing HTCondor backend.")
        # Set initial directory
        if initial_dir is None:
            initial_dir = Path.cwd()
        # Set shared data directory
        if shared_data_dir is None:
            shared_data_dir = Path.cwd() / "joblib_htcondor_shared_data"
        else:
            # Pathify string path
            if isinstance(shared_data_dir, str):
                shared_data_dir = Path(shared_data_dir)
        # Set Python executable
        if python_path is None:
            python_path = sys.executable
        # Make shared data directory if doesn't exist
        shared_data_dir.mkdir(exist_ok=True, parents=True)

        # condor_submit stuff
        self._pool = pool
        self._schedd = schedd
        self._universe = universe
        self._python_path = python_path
        self._request_cpus = request_cpus
        self._request_memory = request_memory
        self._request_disk = request_disk
        self._initial_dir = initial_dir
        self._log_dir_prefix = log_dir_prefix
        self._poll_interval = poll_interval
        self._shared_data_dir = shared_data_dir
        self._extra_directives = extra_directives
        self._worker_log_level = worker_log_level
        self._batch_size = batch_size
        self._max_recursion_level = max_recursion_level
        self._export_metadata = export_metadata

        self._recursion_level = 0
        self._parent_uuid = None

        # Check and set throttle
        if isinstance(throttle, list):
            if len(throttle) == 1:
                self._throttle = throttle[0]
            elif len(throttle) == 0:
                raise ValueError(
                    "Throttle parameter must have at least one value."
                )
        self._throttle = throttle

        logger.debug(f"Universe: {self._universe}")
        logger.debug(f"Python path: {self._python_path}")
        logger.debug(f"Request CPUs: {self._request_cpus}")
        logger.debug(f"Request memory: {self._request_memory}")
        logger.debug(f"Request disk: {self._request_disk}")
        logger.debug(f"Initial dir: {self._initial_dir}")
        logger.debug(f"Log dir prefix: {self._log_dir_prefix}")
        logger.debug(f"Poll interval: {self._poll_interval}")
        logger.debug(f"Shared data dir: {self._shared_data_dir}")
        logger.debug(f"Extra directives: {self._extra_directives}")
        logger.debug(f"Worker log level: {self._worker_log_level}")
        logger.debug(f"Throttle: {self._throttle}")
        logger.debug(f"Batch Size: {self._batch_size}")
        logger.debug(f"Max recursion level: {self._max_recursion_level}")
        logger.debug(f"Export metadata: {self._export_metadata}")

        logger.debug(f"Recursion level: {self._recursion_level}")
        logger.debug(f"Parent UUID: {self._parent_uuid}")

        # Create new scheduler client
        if schedd is None:
            # Try to create a scheduler client using local daemon
            try:
                schedd = htcondor2.Schedd()
            except htcondor2.HTCondorException as err:
                # Initiate collector client
                collector = htcondor2.Collector(self._pool)
                # Query for scheduler ads
                schedd_ads = collector.query(
                    ad_type=htcondor2.AdType.Schedd,
                    projection=["Name", "MyAddress", "CondorVersion"],
                )
                # Sanity check for scheduler ads
                if not schedd_ads:
                    raise RuntimeError(
                        "Unable to locate local daemon."
                    ) from err
                # Create scheduler client
                schedd = htcondor2.Schedd(schedd_ads[0])

        self._client = schedd

        # Create placeholder for polling thread executor, initialized in
        # start_call() and stopped in stop_call()
        self._polling_thread_executor: Optional[ThreadPoolExecutor] = None

        # Create tracking capabilities for queued, waiting and completed jobs
        self._queued_jobs_list: deque[_HTCondorJobMeta] = deque()
        self._waiting_jobs_deque: deque[_HTCondorJobMeta] = deque()
        self._completed_jobs_list: list[_HTCondorJobMeta] = []

        self._n_jobs = 1
        self._backend_meta: Union[_BackendMeta, None] = None

        # Set some initial values for job scheduling
        self._current_shared_data_dir = Path()
        self._next_task_id = 0
        self._this_batch_name = "missingbatchname"
        logger.debug("HTCondor backend initialised.")

    def write_metadata(self) -> None:
        """Write metadata to a file.

        Raises
        ------
        OSError
            If metadata could not be written.

        """
        # Early return if no metadata yet
        if self._backend_meta is None:
            return
        # Update metadata information before dumping
        self._backend_meta.update_timestamp = datetime.now()
        meta_dir = self._shared_data_dir / ".jht-meta"
        meta_dir.mkdir(exist_ok=True, parents=True)
        meta_fname = meta_dir / f"{self._this_batch_name}.json"
        try:
            with meta_fname.open("w") as f:
                json.dump(self._backend_meta.asdict(), f)
        except OSError as e:
            logger.warning(f"Error writing metadata (will retry): {e}")

    def get_nested_backend(self) -> tuple["ParallelBackendBase", int]:
        """Return the nested backend and the number of workers.

        Returns
        -------
        ParallelBackendBase
            The nested backend.
        int
            The number of jobs.

        """
        if self._recursion_level == self._max_recursion_level:
            logger.debug(
                "Maximum recursion level reached. Switching to Sequential "
                "backend."
            )
            return SequentialBackend(
                nesting_level=self._recursion_level
            ), self._n_jobs

        if self._poll_interval < 1:
            logger.warning(
                "You are about to use nested parallel calls. "
                "Poll interval is less than 1 second. This could lead to "
                "high load on the filesystem and/or scheduler. Please "
                "consider increasing the poll interval to a few seconds "
                "as this will not have any considerable effect on the "
                "throughput."
            )
        throttle = self._throttle
        if not isinstance(throttle, list):
            logger.warning(
                "You are about to use nested parallel calls and the throttle "
                "parameter you set would not have the desired effect. "
                "In nested parallel calls, each worker in the parent parallel"
                "will throttle the number of jobs submitted to the child "
                "with the same settings. For example, if you set "
                "`throttle=10` in the parent, then each child will submit 10 "
                "jobs at a time, leading to a total of 110 jobs submitted at "
                "once, which is far from the selected value of 10. In order "
                "to manipulate the thottle value for each level, you need to "
                "set throttle as a list of integers."
            )
        elif len(throttle) > 1:
            throttle = throttle[1:]

        return _HTCondorBackendFactory.build(
            request_cpus=self._request_cpus,
            request_memory=self._request_memory,
            pool=self._pool,
            schedd=self._schedd,
            universe=self._universe,
            python_path=self._python_path,
            request_disk=self._request_disk,
            initial_dir=self._initial_dir,
            log_dir_prefix=self._log_dir_prefix,
            poll_interval=self._poll_interval,
            shared_data_dir=self._shared_data_dir,
            extra_directives=self._extra_directives,
            worker_log_level=self._worker_log_level,
            throttle=throttle,
            batch_size=self._batch_size,
            max_recursion_level=self._max_recursion_level,
            export_metadata=self._export_metadata,
            recursion_level=self._recursion_level + 1,
            parent_uuid=self._this_batch_name,
        ), self._n_jobs

    def __reduce__(self) -> tuple[Callable, tuple]:
        return (
            _HTCondorBackendFactory.build,
            (
                self._request_cpus,
                self._request_memory,
                self._pool,
                self._schedd,
                self._universe,
                self._python_path,
                self._request_disk,
                self._initial_dir,
                self._log_dir_prefix,
                self._poll_interval,
                self._shared_data_dir,
                self._extra_directives,
                self._worker_log_level,
                self._throttle,
                self._batch_size,
                self._max_recursion_level,
                self._export_metadata,
                self._recursion_level,
                self._parent_uuid,
            ),
        )

    def effective_n_jobs(self, n_jobs: int) -> int:
        """Guesstimate of actual jobs.

        Parameters
        ----------
        n_jobs : int
            Theoretical number of jobs.

        Returns
        -------
        int
            Actual number of jobs.

        """
        if n_jobs == -1:
            if self._throttle is None:
                logger.warning(
                    "Setting HTCondor backend n_jobs to -1 does not make much "
                    "sense. We cannot determine the effective number of jobs "
                    "as HTCondor is actually a queuing system. Ideally, you "
                    "should set n_jobs to the number of total jobs you will "
                    "queue, or an arbitrary large number.\n"
                    "IMPORTANT: the number of jobs that joblib with submit "
                    "depends on the pre_dispatch parameter of Parallel, which "
                    "is usually a function of n_jobs. You need to consider "
                    "that the data transfer is done through the filesystem, "
                    "so an extremely large number of jobs with large data "
                    "could lead to a large shared storage requirement. "
                    "If you set n_jobs to a fixed number, joblib will submit "
                    "in batches of that number of jobs, so you won't exploit "
                    "the full potential of this backend. What you might want "
                    "is to throttle the number of jobs submitted at once, "
                    "which can be done with the throttle parameter of the "
                    "backend. As a sane default, we will set the throttle to "
                    "1000 and the number of jobs as the maximum possible "
                    "number of jobs."
                )
                self._throttle = 1000
            n_jobs = sys.maxsize // 2
            logger.info(
                f"Setting n_jobs to {n_jobs} but "
                f"throttling to {self._throttle}"
            )
        else:
            logger.info(
                f"Setting n_jobs to {n_jobs}, though this value "
                "does not have any effect."
            )
        return n_jobs

    def configure(
        self,
        n_jobs: int = 1,
        parallel: Optional[type["Parallel"]] = None,
        **backend_args: Any,
    ) -> int:
        """Reconfigure the backend and return the number of workers.

        It's called inside joblib.Parallel.__call__() quite early on so should
        set correct self._n_jobs for use.

        Parameters
        ----------
        n_jobs : int, optional
            Number of jobs (default 1).
        parallel : joblib.Parallel instance or None, optional
            The joblib.Parallel instance used (default None).
        **backend_args : any
            Keyword arguments to pass to the backend.

        """
        logger.debug("Configuring HTCondor backend.")
        logger.debug(f"\tn_jobs: {n_jobs}")
        logger.debug(f"\tparallel: {parallel}")
        logger.debug(f"\tbackend_args: {backend_args}")
        self._n_jobs = self.effective_n_jobs(n_jobs)
        self.parallel = parallel

        # Create unique id for job batch
        self._uuid = uuid1().hex
        logger.debug(f"\tUUID: {self._uuid}")
        self._this_batch_name = f"jht-{self._uuid}-l{self._recursion_level}"
        logger.info(f"\tBatch name: {self._this_batch_name}")
        logger.debug("HTCondor backend configured.")

        if self._log_dir_prefix is None:
            self._current_log_dir_prefix = (
                f"{self._initial_dir}/logs/{self._this_batch_name}"
            )
        else:
            self._current_log_dir_prefix = self._log_dir_prefix

        logger.info(
            f"Setting log dir prefix to {self._current_log_dir_prefix}"
        )
        log_path = Path(self._current_log_dir_prefix)
        if not log_path.exists() and log_path.is_absolute():
            log_path.mkdir(parents=True)

        return self._n_jobs

    # def compute_batch_size(self) -> int:
    #     """Compute the batch size.

    #     Returns
    #     -------
    #     int
    #         The batch size.

    #     """
    #     logger.debug(f"Computing batch size = {self._batch_size}.")
    #     return self._batch_size

    def apply_async(
        self, func: "BatchedCalls", callback: Optional[Callable] = None
    ) -> type["AsyncResult"]:
        """Call ``func`` and if provided ``callback`` after ``func`` runs.

        It's called inside joblib.Parallel._dispatch() .

        Parameters
        ----------
        func : joblib.parallel.BatchedCalls
            Batched functions to run.
        callback : callable or None, optional
            Callback to run after ``func`` is complete (default None).

        Returns
        -------
        multiprocess.pool.AsyncResult like object
            The object to update result of calling ``func`` asynchronously.

        """
        # Create Future to provide completion info
        f = Future()
        # Match multiprocessing.pool.AsyncResult
        f.get = f.result  # type: ignore
        # Submit to queue
        self._submit(f, func=func, callback=callback)
        # Write metadata
        if self._export_metadata:
            self.write_metadata()
        return f  # type: ignore

    def _submit(
        self,
        submission_future: Future,
        func: "BatchedCalls",
        callback: Optional[Callable] = None,
    ) -> None:
        """Submit a HTCondor job and adds to waiting jobs deque.

        Parameters
        ----------
        submission_future : concurrent.futures.Future
            The Future to attach to job submission.
        func : joblib.parallel.BatchedCalls
            The function to submit.
        callback : callable or None, optional
            The callback to run after the function is complete (default None).

        """
        # Pickle the function to a file
        this_task_id = self._next_task_id
        pickle_fname = (
            self._current_shared_data_dir / f"task-{this_task_id}.pickle"
        )
        self._next_task_id += 1

        # Create the DelayedSubmission object
        ds = DelayedSubmission(func)

        arguments = (
            "-m joblib_htcondor.executor "
            f"--verbose {self._worker_log_level} "
            f"{pickle_fname.as_posix()}"
        )
        # Creat the job submission dictionary
        submit_dict = {
            "universe": self._universe,
            "executable": f"{self._python_path}",
            "arguments": arguments,
            "request_cpus": self._request_cpus,
            "request_memory": self._request_memory,
            "request_disk": self._request_disk,
            "initial_dir": self._initial_dir,
            "transfer_executable": "False",
            "log": (
                f"{self._current_log_dir_prefix}/"
                "$(ClusterId).$(ProcId).joblib.log"
            ),
            "output": (
                f"{self._current_log_dir_prefix}/"
                "$(ClusterId).$(ProcId).joblib.out"
            ),
            "error": (
                f"{self._current_log_dir_prefix}/"
                "$(ClusterId).$(ProcId).joblib.err"
            ),
            "+JobBatchName": f'"jht-{self._uuid}"',
            "+JobPrio": f"{self._recursion_level}",
        }
        if self._extra_directives:
            submit_dict.update(self._extra_directives)

        logger.log(level=9, msg=f"Submit dict: {submit_dict}")
        # Createthe Submit to htcondor
        submit = htcondor2.Submit(submit_dict)

        if self._export_metadata:
            # add a new task to the metadata
            self._backend_meta.task_status.append(  # type: ignore
                _TaskMeta(request_cpus=self._request_cpus)
            )
            self._backend_meta.n_tasks += 1  # type: ignore

        # Add to queue so the poller can take care of it
        self._queued_jobs_list.append(
            _HTCondorJobMeta(
                tracking_future=submission_future,
                htcondor_submit=submit,
                htcondor_submit_result=None,
                callback=callback,
                pickle_fname=pickle_fname,
                delayed_submission=ds,
                task_id=this_task_id,
            )
        )

    def get_current_throttle(self) -> int:
        """Get the current throttle value.

        Returns
        -------
        int
            The current throttle value.

        """
        out = sys.maxsize
        if isinstance(self._throttle, int):
            out = self._throttle
        elif isinstance(self._throttle, list):
            out = self._throttle[0]
        return out

    def _watcher(self) -> None:
        """Long poller for job tracking.

        Parameters
        ----------
        future : concurrent.futures.Future
            The future to set after polling.
        submit_result : htcondor2.SubmitResult
            The result submission object to query with for polling.

        """
        last_poll = time.time()
        throttle = self.get_current_throttle()
        logger.info("Starting HTCondor backend watcher.")
        logger.info(f"Polling every {self._poll_interval} seconds.")
        logger.info(f"Throttle set to {throttle}.")
        while self._continue:
            try:
                if (time.time() - last_poll) > self._poll_interval:
                    # Enough time passed, poll the jobs
                    n_running, update_meta = self._poll_jobs()
                    last_poll = time.time()
                    if n_running < throttle:
                        # We don't have enough jobs int he condor queue, submit
                        newly_queued = 0
                        max_to_queue = throttle - n_running
                        # 1) check if there are any queued jobs to be submitted
                        while (
                            self._queued_jobs_list
                            and newly_queued < max_to_queue
                        ):
                            to_submit = self._queued_jobs_list.popleft()
                            logger.log(
                                level=9, msg=f"Dumping pickle file {to_submit}"
                            )
                            # Dump pickle file
                            dumped = to_submit.delayed_submission.dump(
                                to_submit.pickle_fname
                            )
                            if not dumped:
                                # Something went wrong
                                # continue and submit this later
                                logger.debug("Could not dump pickle file.")
                                continue
                            # Submit job
                            logger.log(
                                level=9, msg=f"Submitting job {to_submit}"
                            )
                            try:
                                to_submit.htcondor_submit_result = (
                                    self._client.submit(
                                        to_submit.htcondor_submit,
                                        count=1,
                                    )
                                )
                            except OSError as e:
                                # Something went wrong, continue and submit
                                # this later
                                logger.error(f"Error submitting job: {e}")
                                logger.error(traceback.format_exc())
                                logger.error("Will try later.")

                                # Put the job back in the queue
                                self._queued_jobs_list.appendleft(to_submit)

                                # Delete the pickle file
                                to_submit.pickle_fname.unlink()

                                # Wait a bit before trying again
                                time.sleep(1)
                                continue

                            logger.log(level=9, msg="Getting cluster id.")
                            # Set the cluster id
                            to_submit.cluster_id = (  # type: ignore
                                to_submit.htcondor_submit_result.cluster()
                            )
                            logger.log(level=9, msg="Job submitted.")
                            # Move to waiting jobs
                            self._waiting_jobs_deque.append(to_submit)
                            newly_queued += 1
                            update_meta = True

                            if self._export_metadata:
                                # Update the sent timestamp and cluster id
                                logger.log(
                                    level=9,
                                    msg="Updating task status timestamp.",
                                )
                                self._backend_meta.task_status[  # type: ignore
                                    to_submit.task_id - 1
                                ].sent_timestamp = datetime.now()

                                logger.log(
                                    level=9,
                                    msg="Updating task status cluster id.",
                                )
                                self._backend_meta.task_status[  # type: ignore
                                    to_submit.task_id - 1
                                ].cluster_id = to_submit.cluster_id

                                logger.log(level=9, msg="Task status updated")

                    if update_meta and self._export_metadata:
                        self.write_metadata()
                # logger.debug("Waiting 0.1 seconds")
                time.sleep(0.1)
            except Exception as e:  # noqa: BLE001
                logger.error(f"Error in HTCondor backend watcher: {e}")
                logger.error(traceback.format_exc())

    def _poll_jobs(self) -> tuple[int, bool]:  # noqa: C901
        """Poll the schedd for job status.

        Returns
        -------
        int
            number of jobs that are queued or running.

        """
        logger.debug("Polling HTCondor jobs.")
        update_meta = False

        if (
            self._next_task_id > 1  # at least one job was queued
            and len(self._waiting_jobs_deque) == 0  # nothing waiting
            and len(self._queued_jobs_list) == 0  # nothing scheduled
        ):
            # No jobs to track, stop polling
            logger.debug("No jobs to track, stopping polling.")
            self._continue = False
        else:
            done_jobs = []
            # Query output files
            for job_meta in self._waiting_jobs_deque:
                logger.log(level=9, msg=f"Checking job {job_meta}")

                out_fname = job_meta.pickle_fname.with_stem(
                    f"{job_meta.pickle_fname.stem}_out"
                )
                if out_fname.exists():
                    logger.log(level=9, msg=f"Job {job_meta} is done.")
                    out_fname = job_meta.pickle_fname.with_stem(
                        f"{job_meta.pickle_fname.stem}_out"
                    )
                    # Load the DelayedSubmission object
                    ds = DelayedSubmission.load(out_fname)
                    if ds is None:
                        # Something went wrong, continue and poll later
                        continue
                    done_jobs.append(job_meta)
                    result = ds.result()
                    if ds.error():
                        logger.log(
                            level=9,
                            msg=f"Job {job_meta} raised an exception.",
                        )
                        job_meta.tracking_future.set_exception(result)
                    else:
                        logger.log(
                            level=9,
                            msg=f"Job {job_meta} completed successfully.",
                        )
                        logger.log(level=9, msg=f"Result: {result}")
                        job_meta.tracking_future.set_result(result)
                        if job_meta.callback:
                            job_meta.callback(result)

                        # Add to completed list
                        self._completed_jobs_list.append(job_meta)
                        if self._export_metadata:
                            # Set the done timestamp from the ds object
                            self._backend_meta.task_status[  # type: ignore
                                job_meta.task_id - 1
                            ].done_timestamp = ds.done_timestamp()

                if self._export_metadata:
                    # After checking for the output file, update the run
                    # timestamp from the run file if needed. This is done after
                    # to ensure that even if the job is done, the run file is
                    # still updated.
                    run_fname = job_meta.pickle_fname.with_suffix(".run")
                    update_meta = (
                        update_meta
                        or self._backend_meta.task_status[  # type: ignore
                            job_meta.task_id - 1
                        ].update_run_from_file(run_fname)
                    )

            # Remove completed jobs
            if len(done_jobs) > 0:
                update_meta = True
                logger.log(
                    level=9,
                    msg=f"Removing jobs from waiting list: {done_jobs}",
                )
                for job_meta in done_jobs:
                    # Free up resources
                    job_meta.pickle_fname.unlink()
                    run_fname = job_meta.pickle_fname.with_suffix(".run")
                    if self._export_metadata:
                        # Update run from file again
                        self._backend_meta.task_status[  # type: ignore
                            job_meta.task_id - 1
                        ].update_run_from_file(run_fname)
                    run_fname.unlink()
                    out_fname = job_meta.pickle_fname.with_stem(
                        f"{job_meta.pickle_fname.stem}_out"
                    )
                    out_fname.unlink()
                    self._waiting_jobs_deque.remove(job_meta)

        logger.debug("Polling HTCondor jobs done.")
        return len(self._waiting_jobs_deque), update_meta

    def start_call(self) -> None:
        """Start resources before actual computation."""
        logger.debug("Starting HTCondor backend.")
        # Set flag for joblib
        self._continue = True
        # Start polling thread executor
        self._polling_thread_executor = ThreadPoolExecutor(
            max_workers=1,
            thread_name_prefix="schedd_poll",
        )
        # Create shared data directory
        self._current_shared_data_dir = (
            self._shared_data_dir / f"{self._this_batch_name}"
        )
        logger.info(
            f"Setting shared data dir to {self._current_shared_data_dir}"
        )

        self._current_shared_data_dir.mkdir(exist_ok=True, parents=True)
        # Set task ID for pickle file name
        self._next_task_id = 1

        if self._export_metadata:
            # Create metadata object
            meta = _BackendMeta(
                uuid=self._this_batch_name,
                parent=self._parent_uuid,
                recursion_level=self._recursion_level,
                throttle=self.get_current_throttle(),
                shared_data_dir=self._current_shared_data_dir,
                n_tasks=0,
            )
            self._backend_meta = meta
            # Write metadata file
            self.write_metadata()

        # Register custom handler for SIGTERM;
        # If we are cancelled, stop the backend so all the jobs are cancelled
        signal.signal(signal.SIGTERM, self._sigterm_handler)

        # Initialize long polling
        self._polling_thread_executor.submit(self._watcher)

    def stop_call(self) -> None:
        """Stop resources after actual computation."""
        logger.debug("Stopping HTCondor backend.")
        # Cleanup
        self._sigterm_handler(None, None)

    def _sigterm_handler(self, signum, stackframe) -> None:
        """Handle SIGTERM."""
        # Set flag for joblib
        logger.info("Received SIGTERM, stopping HTCondor backend.")
        self._continue = False
        # Shutdown polling thread executor
        self._polling_thread_executor.shutdown()  # type: ignore
        # Cancel unfinished (idle, on-hold or running ) jobs
        self._cancel_jobs()
        # Remove shared data directory
        shutil.rmtree(self._current_shared_data_dir)

    def _cancel_jobs(self) -> None:
        """Cancel idle, on-hold or running jobs."""
        logger.debug("Cancelling HTCondor jobs.")
        # Query schedd
        query_result = [
            f"{result.lookup('ClusterId')}.{result.lookup('ProcId')}"
            for result in self._client.query(
                constraint=f'JobBatchName =?= "{self._this_batch_name}"',
                projection=["ClusterId", "ProcId"],
            )
        ]
        # Cancel jobs
        if len(query_result) > 0:
            logger.debug(f"Cancelling: {query_result}")
            _ = self._client.act(
                action=htcondor2.JobAction.Remove,
                job_spec=query_result,
                reason="Cancelled by htcondor_joblib",
            )
            logger.debug("HTCondor jobs cancelled.")


class _HTCondorBackendFactory:
    @staticmethod
    def build(
        request_cpus: int,
        request_memory: str,
        pool: Union[str, "ClassAd", list[str], None] = None,
        schedd: Optional[htcondor2.Schedd] = None,
        universe: str = "vanilla",
        python_path: Optional[str] = None,
        request_disk: str = "0GB",
        initial_dir: Union[str, Path, None] = None,
        log_dir_prefix: Optional[str] = None,
        poll_interval: int = 5,
        shared_data_dir: Union[str, Path, None] = None,
        extra_directives: Optional[dict] = None,
        worker_log_level: int = logging.INFO,
        throttle: Union[int, list[int], None] = None,
        batch_size: int = 1,
        max_recursion_level: int = -1,
        export_metadata: bool = False,
        recursion_level: int = 0,
        parent_uuid: Optional[str] = None,
    ) -> _HTCondorBackend:
        """Build HTCondor backend with extra private attributes.

        Parameters
        ----------
        pool : str, classad2.ClassAd, list of str or None, optional
            Pool to initiate htcondor2.Collector client with (default None).
        schedd : htcondor2.Schedd or None, optional
            Scheduler to use for submitting jobs (default None).
        universe : str, optional
            HTCondor universe to use (default "vanilla").
        python_path : str, optional
            Path to the Python binary (defaults to current python executable).
        request_cpus : int, optional
            HTCondor CPUs to request (default 1).
        request_memory : str, optional
            HTCondor memory to request (default "8GB").
        request_disk : str, optional
            HTCondor disk to request (default "8G").
        initial_dir : str or pathlib.Path or None, optional
            HTCondor initial directory for job. If None, will resolve to
            current working directory (default None).
        log_dir_prefix : str or None, optional
            Prefix for the log directory. The directory prefix needs
            to exist before submitting as HTCondor demands. If None, will
            resolve to `"`initial_dir`/logs/<generated unique batch name>"`
            (default None).
        poll_interval : int, optional
            Interval in seconds to poll the scheduler for job status
            (default 5).
        shared_data_dir : str or pathlib.Path or None, optional
            Directory to store shared data between jobs. If None, will resolve
            to `"<current working directory>/joblib_htcondor_shared_data"`
            (default None).
        extra_directives : dict or None, optional
            Extra directives to pass to the HTCondor submit file
            (default None).
        worker_log_level : int, optional
            Log level for the worker (default is logger.INFO).
        throttle : int or list of int or None, optional
            Throttle the number of jobs submitted at once. If list, the first
            element is the throttle for the current level and the rest are
            for the nested levels (default None).
        batch_size : int, optional
            Number of joblib jobs to group together in a single HTCondor job (
            default 1).
        export_metadata : bool, optional
            Export metadata to a file, to be used with the HTCondor Joblib
            Monitor. This increases the load on the filesystem considerably if
            the number of jobs is high and the duration is short
            (default False).
        recursion_level : int, optional
            Recursion level of the backend. With each nested
            call, the recursion level increases by 1 (default 0).
        max_recursion_level : int, optional
            Maximum recursion level of the backend. Once the
            recursion level reaches this value, the backend will switch to a
            Sequential backend. If -1, the switching is disabled. If 0,
            no recursion is allowed (default 0).
        parent_uuid : str or None, optional
            UUID of the parent backend (default None).

        Returns
        -------
        _HTCondorBackend
            The HTCondor backend instance.

        """
        out = _HTCondorBackend(
            request_cpus=request_cpus,
            request_memory=request_memory,
            pool=pool,
            schedd=schedd,
            universe=universe,
            python_path=python_path,
            request_disk=request_disk,
            initial_dir=initial_dir,
            log_dir_prefix=log_dir_prefix,
            poll_interval=poll_interval,
            shared_data_dir=shared_data_dir,
            extra_directives=extra_directives,
            worker_log_level=worker_log_level,
            throttle=throttle,
            batch_size=batch_size,
            max_recursion_level=max_recursion_level,
            export_metadata=export_metadata,
        )
        out._recursion_level = recursion_level
        out._parent_uuid = parent_uuid  # type: ignore
        return out



"""Joblib HTCondor Backend."""

# Authors: Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

__all__ = ["register_htcondor"]


from typing import Union

from ._version import __version__


def register_htcondor(level: Union[int, str] = "WARNING") -> None:
    """Register htcondor backend into joblib.

    Parameters
    ----------
    level : int or {"DEBUG", "INFO", "WARNING", "ERROR"}
        The level of the messages to print. If string, it will be interpreted
        as elements of logging (default "WARNING").
    """
    from .backend import register
    from .logging import configure_logging
    configure_logging(level=level)
    register()



"""The joblib htcondor UI module entrypoint."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import argparse
import curses
import logging
import traceback
from pathlib import Path
from typing import Union

from .config import (
    COLOR_DONE,
    COLOR_QUEUED,
    COLOR_RUNNING,
    COLOR_SENT,
    PBAR_CHAR,
)
from .main_window import MainWindow
from .uilogging import init_logging, logger
from .utils import (
    align_text,
)


mainwin: Union[MainWindow, None] = None


def color_test():
    """Test color."""
    mainwin.win.clear()  # type: ignore
    line = 0
    col = 0
    for i in range(0, 255):
        col = i % 10 + 3
        line = i // 10
        if col == 3:
            mainwin.win.addstr(line, 0, f"{line}", curses.color_pair(0))
        mainwin.win.addstr(line, col, "■", curses.color_pair(i))
    line = line + 1
    mainwin.win.addstr(line, 0, PBAR_CHAR * 10, curses.color_pair(COLOR_DONE))
    mainwin.win.addstr(
        line, 10, PBAR_CHAR * 15, curses.color_pair(COLOR_RUNNING)
    )
    mainwin.win.addstr(line, 25, PBAR_CHAR * 30, curses.color_pair(COLOR_SENT))
    mainwin.win.addstr(
        line, 55, PBAR_CHAR * 10, curses.color_pair(COLOR_QUEUED)
    )

    mainwin.win.refresh()

    while 1:
        c = mainwin.win.getch()
        if c == ord("q"):
            break


def align_test():
    """Test alignment."""
    mainwin.win.clear()

    align_text(mainwin.win, "Top left", 0, 0, curses.color_pair(5))
    align_text(mainwin.win, "Top Center", 0, "center", curses.color_pair(5))
    align_text(mainwin.win, "Top Right", 0, -1, curses.color_pair(5))

    align_text(mainwin.win, "Middle left", "center", 0, curses.color_pair(5))
    align_text(
        mainwin.win, "Middle Center", "center", "center", curses.color_pair(5)
    )
    align_text(mainwin.win, "Middle Right", "center", -1, curses.color_pair(5))

    align_text(mainwin.win, "Bottom left", -1, 0, curses.color_pair(5))
    align_text(
        mainwin.win, "Bottom Center", -1, "center", curses.color_pair(5)
    )
    align_text(mainwin.win, "Bottom Right", -1, -1, curses.color_pair(5))

    mainwin.win.refresh()

    while 1:
        c = mainwin.win.getch()
        if c == ord("q"):
            break


def main_ui(stdscr, args):
    """Set main UI."""
    global mainwin
    global curdir
    logger.info("==== Starting UI ====")
    logger.debug(f"Can change color: {curses.can_change_color()}")
    logger.debug(f"Has color: {curses.has_colors()}")
    logger.debug(f"Has extended color: {curses.has_extended_color_support()}")
    logger.info(f"Path: {args.path}")

    n_colors = curses.COLORS
    curses.use_default_colors()
    logger.debug(f"Number of colors: {n_colors}")
    for i in range(0, curses.COLORS - 4):
        curses.init_pair(i, i, -1)

    curses.curs_set(0)
    # curses.init_pair(0, curses.COLOR_WHITE, -1)
    # curses.init_pair(1, curses.COLOR_RED, -1)
    # curses.init_pair(2, curses.COLOR_GREEN, -1)
    # curses.init_pair(3, curses.COLOR_BLUE, -1)

    # center_text(stdscr, "HTCondor Joblib Monitor", 0, curses.color_pair(5))
    # stdscr.addstr(1,0, "RED ALERT!", curses.color_pair(5))
    if args.test == "color":
        mainwin = MainWindow(stdscr, args.path)
        color_test()
    elif args.test == "align":
        mainwin = MainWindow(stdscr, args.path)
        align_test()
    else:
        try:
            mainwin = MainWindow(stdscr, args.path, args.refresh)
            mainwin.render()
            mainwin.event_handler()
        except curses.error as e:
            logger.error(traceback.format_exc())
            logger.error(e)


if __name__ == "__main__":

    def dir_or_file_path(string):
        """Handle directory or file and return path."""
        t_path = Path(string)
        if t_path.is_file():
            return t_path
        elif t_path.is_dir():
            t_path = t_path / ".jht-meta"
            if t_path.exists():
                return t_path
        raise ValueError(f"This is not a shared data dir: {string}")

    parser = argparse.ArgumentParser(description="HTCondor Joblib Monitor")
    parser.add_argument(
        "--test",
        help="Run the specified test.",
        choices=["color", "align"],
    )
    parser.add_argument(
        "--refresh",
        type=int,
        default=5,
        help="The refresh interval in seconds.",
    )
    parser.add_argument(
        "--path",
        type=dir_or_file_path,
        required=True,
        help="Directory to look for meta files or a specific meta file.",
    )
    parser.add_argument(
        "--verbose",
        type=int,
        default=logging.INFO,
        help="The logging verbosity to use.",
    )
    args = parser.parse_args()
    init_logging(args.verbose)
    curses.wrapper(main_ui, args)



"""The joblib htcondor DelayedSubmission implementation."""

# Authors: Synchon Mandal <s.mandal@fz-juelich.de>
#          Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

from concurrent.futures.process import _ExceptionWithTraceback
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Optional, Union

from flufl.lock import Lock, TimeOutError  # type: ignore
from joblib.externals.cloudpickle import cloudpickle  # type: ignore

from .logging import logger


__all__ = ["DelayedSubmission"]


def _get_lock(fname: Union[Path, str], *args: Any, **kwargs: Any) -> Lock:
    """Get a `flufl.lock.Lock` object.

    Parameters
    ----------
    fname : pathlib.Path or str
        The lockfile path.
    *args
        Positional arguments passed to `flufl.lock.Lock`.
    **kwargs
        Keyword arguments passed to `flufl.lock.Lock`.

    Returns
    -------
    flufl.lock.Lock
        Lock object.

    """
    if not isinstance(fname, Path):
        fname = Path(fname)
    lock_fname = fname.with_suffix(".lock")

    return Lock(lock_fname.as_posix(), *args, **kwargs)


class DelayedSubmission:
    """Delayed submission object to be run in the worker.

    Implements an object that wraps a function call and its arguments so they
    can be pickled and executed in the workers.

    Parameters
    ----------
    func : callable
        The function to call.
    *args
        Positional arguments to pass to the function.
    **kwargs
        Keyword arguments to pass to the function.

    """

    def __init__(
        self,
        func: Callable,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs
        # Initialize tracking variables
        self._result = None
        self._done = False
        self._error = False
        self._done_timestamp = None

    def run(self) -> None:
        """Run the function with the arguments and store the result."""
        try:
            self._result = self.func(*self.args, **self.kwargs)  # type: ignore
        except BaseException as e:  # noqa: BLE001
            self._result = _ExceptionWithTraceback(
                e,
                e.__traceback__,  # type: ignore
            )
            self._error = True
        self._done_timestamp = datetime.now()
        self._done = True

    def done(self) -> bool:
        """Return whether the function has been run.

        Returns
        -------
        bool
            Whether the function has been run.

        """
        return self._done

    def done_timestamp(self) -> Optional[datetime]:
        """Return the timestamp when the function has finished.

        Returns
        -------
        datetime
            The timestamp when the function has been run. If the function has
            not been run, returns None.

        """
        return self._done_timestamp

    def error(self) -> bool:
        """Return whether the function raised an exception.

        Returns
        -------
        bool
            Whether the function raised an exception.

        """
        return self._error

    def result(self) -> Any:
        """Return the result of the function.

        Returns
        -------
        Any
            The result of the function.

        """
        return self._result

    def dump(
        self, filename: Union[str, Path], result_only: bool = False
    ) -> bool:
        """Dump the object to a file.

        Parameters
        ----------
        filename : str or pathlib.Path
            The file to dump the object to.
        result_only : bool, optional
            Whether to dump only the result (default False).

        """
        if result_only:
            # Avoid pickling function and arguments
            tmp_func = self.func
            tmp_args = self.args
            tmp_kwargs = self.kwargs
            self.func = None
            self.args = None
            self.kwargs = None
        # Get lockfile
        flock = _get_lock(fname=filename, lifetime=120)  # Max 2 minutes
        # Dump in the lockfile
        try:
            with flock:
                with open(filename, "wb") as file:
                    cloudpickle.dump(self, file)
        except TimeOutError:
            logger.error(f"Could not obtain lock for {filename} in 2 minutes.")
            return False
        # Set to original values
        if result_only:
            self.func = tmp_func
            self.args = tmp_args
            self.kwargs = tmp_kwargs

        return True

    @classmethod
    def load(
        cls: type["DelayedSubmission"], filename: Union[str, Path]
    ) -> Optional["DelayedSubmission"]:
        """Load a DelayedSubmission object from a file.

        Parameters
        ----------
        filename : str or pathlib.Path
            The file to load the object from.

        Returns
        -------
        DelayedSubmission or None
            The loaded DelayedSubmission object. If a TimeOutError is raised
            while obtaining the lock, returns None.

        Raises
        ------
        TypeError
            If loaded object is not of type `cls`.

        """
        # Get lockfile
        flock = _get_lock(filename, lifetime=120)  # Max 2 minutes
        # Load from the lockfile
        try:
            with flock:
                with open(filename, "rb") as file:
                    obj = cloudpickle.load(file)
            if not (isinstance(obj, cls)):
                raise TypeError(
                    "Loaded object is not a DelayedSubmission object."
                )
        except TimeOutError:
            logger.error(f"Could not obtain lock for {filename} in 2 minutes.")
            return None
        return obj



"""The joblib htcondor UI logging."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import logging


__all__ = ["init_logging"]


logger = logging.getLogger("joblib_htcondor.ui")


def init_logging(level: int) -> None:
    """Initialise logging for UI.

    Parameters
    ----------
    level : int
        The log level to set.

    """
    # Set logger level
    logger.setLevel(level)
    # Create log file handler
    fh = logging.FileHandler("ui.log")

    output_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    # Create logging formatter
    formatter = logging.Formatter(fmt=output_format)

    fh.setFormatter(formatter)  # set formatter
    # Add log file handler
    logger.addHandler(fh)
    # Set handler level
    fh.setLevel(level)



"""The joblib htcondor UI tree monitor."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import threading
import time
import traceback
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

from .treeparser import MetaTree, parse
from .uilogging import logger


class TreeMonitor:
    """Class for tree monitor.

    Use this class to monitor a tree for changes, regularly updating the tree
    object.

    Parameters
    ----------
    curpath : Path
        The path to the json file to monitor as the root of the tree.
    refresh_interval : int, optional
        The minimum interval in seconds between updates.

    """

    def __init__(self, curpath: Path, refresh_interval: int = 5) -> None:
        self._curpath = curpath
        self._refresh_interval = refresh_interval
        self._last_update = time.time()
        self._curtree = None
        self._lock = threading.Lock()
        self._continue = True
        self._parse_tree()

    def set_path(self, curpath: Path) -> None:
        """Change the path of the tree to monitor.

        curpath : Path
            The path to the json file to monitor as the root of the tree.

        """
        with self._lock:
            self._curpath = curpath
            self._curtree = None
        self._parse_tree()

    def run(self) -> None:
        """Monitor the tree, udpating it regularly."""
        logger.info("Starting tree monitor")
        while self._continue:
            now = time.time()
            if now - self._last_update > self._refresh_interval:
                logger.debug("Checking for updates")
                if self._curpath.is_file():
                    self._parse_tree()
                self._last_update = now
            time.sleep(max(0, 0.1 - (time.time() - now)))

    def _parse_tree(self) -> None:
        """Parse tree for monitor."""
        # logger.debug("P(Lock) for parsing")
        with self._lock:
            parse_from = self._curpath
        # logger.debug("V(Lock) for parsing")

        # Now try to parse an update
        if parse_from.is_dir():
            return
        logger.debug(f"Parsing tree from: {parse_from}")
        try:
            newtree = parse(parse_from)

            # logger.debug("P(Lock) for updating")
            with self._lock:
                self._curtree = newtree
            # logger.debug("V(Lock) for updating")
        except Exception as e:  # noqa: BLE001
            logger.error(
                f"Error parsing tree from {parse_from}: "
                f"{e}\n{traceback.format_exc()}"
            )

    def get_tree(self) -> MetaTree:
        """Get the current tree.

        Returns
        -------
        MetaTree instance
            The current tree.

        """
        # logger.debug("P(Lock) for get_tree")
        with self._lock:
            treecopy = self._curtree
        # logger.debug("V(Lock) for get_tree")
        return treecopy  # type: ignore

    def start(self) -> None:
        """Start monitor."""
        self._monitor_executor = ThreadPoolExecutor(
            max_workers=1,
            thread_name_prefix="ui_treepoller",
        )
        self._monitor_executor.submit(self.run)

    def stop(self) -> None:
        """Stop monitor."""
        if self._monitor_executor is not None:
            self._continue = False
            self._monitor_executor.shutdown(wait=False, cancel_futures=True)
            self._monitor_executor = None



"""The joblib htcondor UI configuration parameters."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

MAX_ENTRIES = 200  # Maximum number of entries to display in the pad

# Colors for the different task states
COLOR_DONE = 56
COLOR_RUNNING = 9
COLOR_SENT = 4
COLOR_QUEUED = 239

# Color thresholds for the disk usage progress bar
COLOR_TRESHOLDS = [56, 4, 13, 9]
COLOR_NONE = 239

# Progress bar character
PBAR_CHAR = "■"



"""The joblib htcondor UI utils."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import curses
from typing import Any, Union

from .config import (
    COLOR_DONE,
    COLOR_QUEUED,
    COLOR_RUNNING,
    COLOR_SENT,
    PBAR_CHAR,
)
from .uilogging import logger


def space_to_unit(size: int) -> tuple[float, str]:
    """Convert size to human readable unit.

    Parameters
    ----------
    size : int
        The size to convert

    Returns
    -------
    tuple[int, str]
        The converted size and unit

    """
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size < 1024:
            return size, unit
        size /= 1024  # type: ignore
    return size, "PB"


def align_text(  # noqa: C901
    screen: Any,
    text: str,
    y: Union[str, int],
    x: Union[str, int],
    *args: Any,
    fill: int = 0,
    underline: int = -1,
) -> tuple[int, int, int]:
    """Align text.

    Parameters
    ----------
    screen : Any
        The screen object to draw on.
    text : str
        The text to draw.
    y : Union[str, int]
        The y position to draw at. It can be an integer or "center".
    x : Union[str, int]
        The x position to draw at. It can be an integer or "center".
    *args : Any
        The arguments to pass to addstr.
    fill : int
        If > 0, fill the text with spaces on both sides to reach this length.
    underline : int
        The character to underline. If -1, do not underline any character.

    Returns
    -------
    y: int
        The y position of the text.
    x: int
        The x position of the text.
    x2: int
        The x position of the last character of the text.

    """
    maxy, maxx = screen.getmaxyx()
    orig_text = text
    if len(args) > 1:
        raise ValueError("Too many arguments")
    if isinstance(y, int):
        if y < 0:
            y = maxy + y
    elif y == "center":
        y = maxy // 2
    else:
        raise ValueError("Invalid y value")

    if isinstance(x, int):
        if fill != 0:
            text = text.center(len(text) + fill)
        if x < 0:
            x = maxx - len(text) + x + 1
    elif x == "center":
        if fill != 0:
            if fill < 0:
                x = -fill
            else:
                x = 0
            text = text.center(maxx - x - 1)
        else:
            x = maxx // 2 - len(text) // 2
    else:
        raise ValueError("Invalid x value")

    try:
        if underline < 0:
            screen.addstr(y, x, text, *args)
        else:
            logger.log(level=9, msg=f"Underline {orig_text} at {underline}")
            t_x = x
            underlined = False
            for ch in text:
                if orig_text[underline] == ch and not underlined:
                    if len(args) == 0:
                        screen.addch(y, t_x, ch, curses.A_UNDERLINE)
                    else:
                        screen.addch(y, t_x, ch, args[0] | curses.A_UNDERLINE)
                    underlined = True
                else:
                    screen.addch(y, t_x, ch, *args)
                t_x += 1  # type: ignore
    except curses.error:
        # This is a hack to avoid the error when we are printing a string that
        # finishes at the very end of the screen.
        pass
    return y, x, x + len(text) - 1  # type: ignore


def table_header(
    screen: Any,
    beginy: int,
    beginx: int,
    length: int,
    text: str,
    align: str = "right",
) -> None:
    """Draw a table header cell.

    Parameters
    ----------
    screen : Any
        The screen object to draw on.
    beginy : int
        The y position to draw at.
    beginx : int
        The x position to draw at.
    length : int
        The length of the header.
    text : str
        The text to draw.
    align : str
        The alignment of the text.

    """
    text_len = len(text)
    if length < text_len + 2:
        raise ValueError("Length too short for table header")
    if text_len == 0:
        for i in range(0, length):
            screen.addch(beginy, beginx + i, curses.ACS_HLINE)
        return
    if align == "left":
        screen.addch(beginy, beginx, curses.ACS_URCORNER)
        screen.addch(beginy, beginx + text_len + 1, curses.ACS_ULCORNER)
        for i in range(2 + text_len, length):
            screen.addch(beginy, beginx + i, curses.ACS_HLINE)
        screen.addstr(beginy, beginx + 1, text)
    elif align == "right":
        for i in range(0, length - text_len - 2):
            screen.addch(beginy, beginx + i, curses.ACS_HLINE)
        screen.addch(
            beginy, beginx + length - text_len - 2, curses.ACS_URCORNER
        )
        screen.addch(beginy, beginx + length - 1, curses.ACS_ULCORNER)
        screen.addstr(beginy, beginx + length - text_len - 1, text)


def table_cell(
    screen: Any,
    beginy: int,
    beginx: int,
    length: int,
    text: str,
    align: str = "right",
):
    """Draw a table cell.

    Parameters
    ----------
    screen : Any
        The screen object to draw on.
    beginy : int
        The y position to draw at.
    beginx : int
        The x position to draw at.
    length : int
        The length of the header.
    text : str
        The text to draw.
    align : str
        The alignment of the text.

    """
    text_len = len(text)
    if length < text_len + 2:
        raise ValueError("Length too short for table cell")
    if align == "right":
        text = text.rjust(length - 2)
    elif align == "center":
        text = text.center(length - 2)
    else:
        text = text.ljust(length - 2)
    screen.addstr(beginy, beginx + 1, text)


def progressbar(
    screen: Any,
    beginy: int,
    beginx: int,
    length: int,
    done: int,
    running: int,
    sent: int,
    queued: int,
) -> None:
    """Draw a batch progress bar.

    Parameters
    ----------
    screen : Any
        The screen object to draw on.
    beginy : int
        The y position to draw at.
    beginx : int
        The x position to draw at.
    length : int
        The size of the progress bar.
    done : int
        The number of done tasks.
    running : int
        The number of running tasks.
    sent : int
        The number of sent tasks.
    queued : int
        The number of queued tasks.

    """
    total = done + running + sent + queued
    if total == 0:
        total = 1
    done_len = round(done / total * length)
    running_len = round(running / total * length)
    submitted_len = round(sent / total * length)
    queued_len = round(queued / total * length)
    all_len = done_len + running_len + submitted_len + queued_len
    while all_len < length:
        if queued_len > 0:
            queued_len += 1
        elif submitted_len > 0:
            submitted_len += 1
        elif running_len > 0:
            running_len += 1
        else:
            done_len += 1
        all_len += 1
    while all_len > length:
        if done_len > 0:
            done_len -= 1
        elif running_len > 0:
            running_len -= 1
        elif submitted_len > 0:
            submitted_len -= 1
        else:
            queued_len -= 1
        all_len -= 1
    logger.log(
        level=9,
        msg=f"Rendering progress bar: {done_len} {running_len} "
        f"{submitted_len} {queued_len}",
    )
    screen.addstr(
        beginy,
        beginx,
        PBAR_CHAR * done_len,
        curses.color_pair(COLOR_DONE),
    )
    screen.addstr(
        beginy,
        beginx + done_len,
        PBAR_CHAR * running_len,
        curses.color_pair(COLOR_RUNNING),
    )
    screen.addstr(
        beginy,
        beginx + done_len + running_len,
        PBAR_CHAR * submitted_len,
        curses.color_pair(COLOR_SENT),
    )
    screen.addstr(
        beginy,
        beginx + done_len + running_len + submitted_len,
        PBAR_CHAR * queued_len,
        curses.color_pair(COLOR_QUEUED),
    )



"""The joblib htcondor UI Main Window."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import curses
import platform
import shutil
import time
from datetime import datetime

from .config import (
    COLOR_DONE,
    COLOR_NONE,
    COLOR_QUEUED,
    COLOR_RUNNING,
    COLOR_SENT,
    COLOR_TRESHOLDS,
    PBAR_CHAR,
)
from .monitor import TreeMonitor
from .open_window import OpenMenu
from .treeparser import MetaTree
from .uilogging import logger
from .utils import (
    align_text,
    progressbar,
    space_to_unit,
    table_cell,
    table_header,
)
from .window import Window


class MainWindow(Window):
    """Class for the main UI window."""

    def __init__(self, window, curpath=None, refresh=5):
        logger.debug("MAIN WINDOW: Initializing")

        self.win = window
        super().__init__(0, 0, window)
        self.subwindows: list[Window] = []
        self.win.attrset(curses.color_pair(5))
        self.curpath = curpath
        self.clear_tree()
        self.treemonitor = TreeMonitor(
            curpath=curpath,  # type: ignore
            refresh_interval=refresh,
        )
        self.treemonitor.start()
        self.refresh_interval = refresh

    def get_meta_dir(self):
        """Get metadata directory."""
        if self.curpath is None:
            return None
        if self.curpath.is_file():
            return self.curpath.parent
        return self.curpath

    def set_path(self, path):
        """Set path for monitor."""
        self.curpath = path
        self.treemonitor.set_path(path)
        self.refresh()

    def clear_tree(self):
        """Clear tree."""
        self.idx_selected = -1
        self.idx_first = 0

    def border(self):
        """Set border."""
        for x in range(1, self.w - 1):
            self.win.addch(0, x, curses.ACS_HLINE)
            self.win.addch(self.h - 1, x, curses.ACS_HLINE)
        for y in range(1, self.h - 1):
            self.win.addch(y, 0, curses.ACS_VLINE)
            self.win.addch(y, self.w - 1, curses.ACS_VLINE)
        self.win.addch(0, 0, curses.ACS_ULCORNER)
        self.win.addch(0, self.w - 1, curses.ACS_URCORNER)
        self.win.addch(self.h - 1, 0, curses.ACS_LLCORNER)
        try:
            self.win.addch(self.h - 1, self.w - 1, curses.ACS_LRCORNER)
        except curses.error:
            pass

    def get_menu(self):
        """Get context menu."""
        if len(self.subwindows) > 0:
            return self.subwindows[-1].get_menu()
        return [(0, "Open")]

    def update_frame(self):
        """Update window frame."""
        logger.log(level=9, msg="MAIN WINDOW: Updating frame")
        try:
            self.win.attrset(curses.color_pair(5))
            self.border()

            align_text(
                self.win,
                PBAR_CHAR + " queued ",
                1,
                -2,
                curses.color_pair(COLOR_QUEUED),
            )
            align_text(
                self.win,
                PBAR_CHAR + " sent",
                1,
                -12,
                curses.color_pair(COLOR_SENT),
            )
            align_text(
                self.win,
                PBAR_CHAR + " running",
                1,
                -19,
                curses.color_pair(COLOR_RUNNING),
            )
            align_text(
                self.win,
                PBAR_CHAR + " done",
                1,
                -29,
                curses.color_pair(COLOR_DONE),
            )
            y, xl, xr = align_text(
                self.win,
                f" {platform.node()} ",
                0,
                3,
            )
            self.win.addch(y, xl - 1, curses.ACS_URCORNER)
            self.win.addch(y, xr + 1, curses.ACS_ULCORNER)
            y, xl, xr = align_text(
                self.win,
                f" {datetime.now().strftime('%a %b %d %H:%M:%S %Y')} ",
                0,
                -3,
            )
            self.win.addch(y, xl - 1, curses.ACS_URCORNER)
            self.win.addch(y, xr + 1, curses.ACS_ULCORNER)
            y, xl, xr = align_text(
                self.win,
                " HTCondor Joblib Monitor ",
                0,
                "center",
            )
            self.win.addch(y, xl - 1, curses.ACS_URCORNER)
            self.win.addch(y, xr + 1, curses.ACS_ULCORNER)

            quit_text = "Quit"
            y, xl, xr = align_text(
                self.win,
                quit_text,
                -1,
                -3,
                fill=2,
                underline=0,
            )
            self.win.addch(y, xl - 1, curses.ACS_URCORNER)
            self.win.addch(y, xr + 1, curses.ACS_ULCORNER)
            curr_x = -3 - len(quit_text) - 4
            menu = self.get_menu()
            for i, t in menu:
                y, xl, xr = align_text(
                    self.win,
                    t,
                    -1,
                    curr_x,
                    fill=2,
                    underline=i,
                )
                self.win.addch(y, xl - 1, curses.ACS_URCORNER)
                self.win.addch(y, xr + 1, curses.ACS_ULCORNER)
                curr_x -= len(t) + 4

        except curses.error:
            # This is a hack to avoid the error when the terminal being resized
            # too fast that the update_frame() is called before the
            # update_lines_cols() is called.
            pass

    def resize(self):
        """Handle window resize."""
        super().resize()
        for win in self.subwindows:
            win.resize()

    def _render_tree_element(
        self,
        tree: MetaTree,
        y: int,
        level: int,
        idx_element: int,
        idx_skip: int,
    ) -> tuple[int, int, int]:
        """Render tree element."""
        if y >= self.h - 2:
            return y, idx_element, idx_skip
        if idx_skip > 0:
            idx_skip -= 1
        else:
            # We are rendering this element
            task_status = tree.get_task_status()
            if idx_element == self.idx_selected:
                self.win.attrset(curses.color_pair(15))
            else:
                if (
                    task_status["total"] > 0
                    and task_status["total"] == task_status["done"]
                ):
                    self.win.attrset(curses.color_pair(COLOR_DONE))

                else:
                    self.win.attrset(curses.color_pair(5))

            uuid_text = tree.meta.uuid[
                : self.batch_field_size - 2 * (level + 1)
            ]
            self.win.addstr(
                y,
                2,
                " " * (level * 2) + uuid_text,
            )

            if level > 0:
                self.win.addch(y, 2 + (level - 1) * 2, curses.ACS_LLCORNER)
                self.win.addch(y, 3 + (level - 1) * 2, curses.ACS_HLINE)

            table_cell(
                self.win,
                y,
                self.batch_field_size,
                8,
                f"{task_status['done']}",
            )
            table_cell(
                self.win,
                y,
                self.batch_field_size + 8,
                8,
                f"{task_status['running']}",
            )
            table_cell(
                self.win,
                y,
                self.batch_field_size + 16,
                8,
                f"{task_status['sent']}",
            )
            table_cell(
                self.win,
                y,
                self.batch_field_size + 24,
                8,
                f"{task_status['queued']}",
            )
            table_cell(
                self.win,
                y,
                self.batch_field_size + 32,
                8,
                f"{task_status['total']}",
            )

            table_cell(
                self.win,
                y,
                self.batch_field_size + 40,
                8,
                f"{tree.meta.throttle}",
            )

            progressbar(
                self.win,
                y,
                self.batch_field_size + 48,
                self.w - self.batch_field_size - 48 - 2,
                task_status["done"],
                task_status["running"],
                task_status["sent"],
                task_status["queued"],
            )
            y += 1
        idx_element += 1
        for c in tree.children:
            y, idx_element, idx_skip = self._render_tree_element(
                c, y, level + 1, idx_element, idx_skip
            )
        return y, idx_element, idx_skip

    def _render_summary_element(self, summary, y_start, title):
        """Render element summary."""
        table_cell(
            self.win,
            y_start,
            1,
            self.batch_field_size,
            title,
            align="left",
        )
        table_cell(
            self.win, y_start, self.batch_field_size, 8, str(summary["done"])
        )
        table_cell(
            self.win,
            y_start,
            self.batch_field_size + 8,
            8,
            str(summary["running"]),
        )
        table_cell(
            self.win,
            y_start,
            self.batch_field_size + 16,
            8,
            str(summary["sent"]),
        )
        table_cell(
            self.win,
            y_start,
            self.batch_field_size + 24,
            8,
            str(summary["queued"]),
        )
        table_cell(
            self.win,
            y_start,
            self.batch_field_size + 32,
            8,
            str(summary["total"]),
        )
        table_cell(
            self.win,
            y_start,
            self.batch_field_size + 40,
            8,
            str(summary["throttle"]),
        )
        progressbar(
            self.win,
            y_start,
            self.batch_field_size + 48,
            self.w - self.batch_field_size - 48 - 2,
            summary["done"],
            summary["running"],
            summary["sent"],
            summary["queued"],
        )
        y_start += 1

    def scroll_up(self):
        """Handle scroll up."""
        if self.idx_selected > 0:
            self.idx_selected = self.idx_selected - 1
            logger.debug(
                f"MAIN WINDOW: Scroll up: selected: {self.idx_selected}"
            )
            if self.idx_selected < self.idx_first:
                self.idx_first -= 1
                logger.debug(f"  Start reached: {self.idx_first}")

    def scroll_down(self):
        """Handle scroll down."""
        curtree = self.treemonitor.get_tree()
        header_size = curtree.depth() + 3
        if self.idx_selected < curtree.size() - 1:
            self.idx_selected = self.idx_selected + 1
            logger.debug(
                f"MAIN WINDOW: Scroll down: selected: {self.idx_selected}"
            )
            if self.idx_selected - self.idx_first >= self.h - header_size - 5:
                self.idx_first += 1
                logger.debug(f"  End reached: {self.idx_first}")

    def render_data(self, y_start=2):
        """Render data."""
        curtree = self.treemonitor.get_tree()
        self.batch_field_size = 50 if self.w > 140 else 20
        n_levels = self.render_summary(y_start)
        self.render_tree(y_start + n_levels + 2)
        self.win.attrset(curses.color_pair(9))
        elapsed = datetime.now() - curtree.meta.start_timestamp  # type: ignore
        days = int(elapsed.total_seconds() // 86400)
        hours = int(elapsed.total_seconds() % 86400) // 3600
        minutes = int(elapsed.total_seconds() % 3600) // 60
        seconds = int(elapsed.total_seconds() % 60)

        # Compute free/used space
        total, used, free = shutil.disk_usage(self.curpath)  # type: ignore

        r_used = used / total
        space_bar_len = 30
        logger.log(level=9, msg=f"Used space ratio: {r_used}")
        color_idx = int((r_used * 4) // 1)
        logger.log(level=9, msg=f"Color index: {color_idx}")
        color = COLOR_TRESHOLDS[color_idx]

        p_used = int(r_used * 100)

        l_used = int(round(r_used * space_bar_len, 1))
        l_free = space_bar_len - l_used
        logger.log(level=9, msg=f"Free space length: {l_free}")

        # Render free/used disk space with units
        u_used, unit_used = space_to_unit(used)
        u_total, unit_total = space_to_unit(total)
        label_free = f"{u_used:.1f} {unit_used} of {u_total:.1f} {unit_total}"
        self.win.addstr(
            self.h - 2,
            self.w - 2 - len(label_free),
            label_free,
            curses.color_pair(9),
        )

        # Render label with % used
        label_used = f"Disk used: {p_used}%"
        self.win.addstr(
            self.h - 2,
            self.w - space_bar_len - 4 - len(label_used) - len(label_free),
            label_used,
            curses.color_pair(9),
        )

        # Render progress bar, first free in grey, then used in color
        self.win.addstr(
            self.h - 2,
            self.w - space_bar_len - 3 + l_used - len(label_free),
            PBAR_CHAR * l_free,
            curses.color_pair(COLOR_NONE),
        )

        self.win.addstr(
            self.h - 2,
            self.w - space_bar_len - 3 - len(label_free),
            PBAR_CHAR * l_used,
            curses.color_pair(color),
        )

        text_elapsed = "Elapsed: "
        if days > 0:
            text_elapsed += f"{days}d "
        text_elapsed += f"{hours:02d}h {minutes:02d}m {seconds:02d}s "
        align_text(
            self.win,
            text_elapsed,
            self.h - 2,
            2,
        )
        # new_x = len(text_elapsed) + 3
        # text_batches = f"- Batches: {treesize}"
        # align_text(
        #     self.win,
        #     text_batches,
        #     self.h - 2,
        #     new_x,
        # )

        ts_update = curtree.last_update()  # type: ignore
        ts_update = ts_update.strftime("%d/%m/%Y %H:%M:%S")
        text_updated = f"- Updated: {ts_update}"

        # new_x = len(text_elapsed) + len(text_batches) + 3
        new_x = len(text_elapsed) + 2
        align_text(
            self.win,
            text_updated,
            self.h - 2,
            new_x,
        )

        if self.idx_first > 0:
            align_text(
                self.win,
                "▲",
                8,
                0,
            )
        if self.idx_first + self.h - n_levels - 5 < curtree.size():
            align_text(
                self.win,
                "▼",
                self.h - 3,
                0,
            )

    def render_summary(self, y_start=2):
        """Render summary."""
        curtree = self.treemonitor.get_tree()
        n_levels = curtree.depth()
        self.win.attrset(curses.color_pair(9))
        table_header(
            self.win,
            y_start,
            1,
            self.batch_field_size,
            "Summary",
            align="left",
        )
        table_header(self.win, y_start, self.batch_field_size, 8, "Done")
        table_header(self.win, y_start, self.batch_field_size + 8, 8, "Run")
        table_header(self.win, y_start, self.batch_field_size + 16, 8, "Sent")
        table_header(
            self.win, y_start, self.batch_field_size + 24, 8, "Queued"
        )
        table_header(self.win, y_start, self.batch_field_size + 32, 8, "Total")
        table_header(
            self.win, y_start, self.batch_field_size + 40, 8, "Throt."
        )
        table_header(
            self.win,
            y_start,
            self.batch_field_size + 48,
            self.w - self.batch_field_size - 48 - 2,
            "",
        )
        status_summary = curtree.get_level_status_summary()  # type: ignore
        y_start += 1
        for i, s in enumerate(status_summary):
            self.win.attrset(curses.color_pair(5))
            self._render_summary_element(
                s,
                y_start,
                title=f"Level {i}" if n_levels > 1 else "Total",
            )
            y_start += 1

        if n_levels > 1:
            total = {
                "done": 0,
                "running": 0,
                "sent": 0,
                "queued": 0,
                "total": 0,
                "throttle": "N/A",
            }
            for s in status_summary:
                for k, v in s.items():
                    if k != "throttle":
                        total[k] += v
            self.win.attrset(curses.color_pair(5))
            self._render_summary_element(total, y_start, title="Total")
            y_start += 1
            n_levels += 1
        return n_levels

    def render_tree(self, y_start=2):
        """Render tree."""
        # logger.debug(f"Tree: {self.curtree}")

        self.win.attrset(curses.color_pair(9))
        table_header(
            self.win,
            y_start,
            1,
            self.batch_field_size,
            "Batch ID",
            align="left",
        )
        table_header(self.win, y_start, self.batch_field_size, 8, "Done")
        table_header(self.win, y_start, self.batch_field_size + 8, 8, "Run")
        table_header(self.win, y_start, self.batch_field_size + 16, 8, "Sent")
        table_header(
            self.win, y_start, self.batch_field_size + 24, 8, "Queued"
        )
        table_header(self.win, y_start, self.batch_field_size + 32, 8, "Total")
        table_header(
            self.win, y_start, self.batch_field_size + 40, 8, "Throt."
        )
        table_header(
            self.win,
            y_start,
            self.batch_field_size + 48,
            self.w - self.batch_field_size - 48 - 2,
            "",
        )
        self.win.attrset(curses.color_pair(5))
        logger.log(level=9, msg=f"Rendering tree starting at {self.idx_first}")
        self._render_tree_element(
            self.treemonitor.get_tree(),  # type: ignore
            y_start + 1,
            0,
            idx_element=0,
            idx_skip=self.idx_first,
        )

    def render(self):
        """Render main window."""
        logger.debug("MAIN WINDOW: Rendering")
        self.win.clear()
        if self.w < 80 or self.h < 20:
            self.win.addstr(
                1,
                1,
                "Window too small. Please resize to at least 80x20",
                curses.color_pair(1),
            )
            self.refresh()
            return
        self.update_frame()

        self.refresh()

    def event_handler(self):
        """Handle event."""
        _continue = True
        self.win.timeout(1000)
        while _continue:
            self.update_frame()
            if (
                len(self.subwindows) == 0
                and (time.time() - self._last_refresh) > self.refresh_interval
            ):
                self.refresh()
            c = self.win.getch()
            logger.log(level=9, msg=f"Key pressed: {c}")
            if c == ord("q"):
                _continue = False  # Exit the while()
                self.treemonitor.stop()
            elif len(self.subwindows) > 0:
                self.subwindows[-1].action(c)
            elif c == ord("o"):
                if len(self.subwindows) == 0:
                    win = OpenMenu(self)
                    self.subwindows.append(win)
                    self.render()
            elif c == curses.KEY_UP:
                self.scroll_up()
                self.refresh()
            elif c == curses.KEY_DOWN:
                self.scroll_down()
                self.refresh()
            elif c == curses.KEY_RESIZE:
                curses.update_lines_cols()
                self.resize()
                self.render()

    def refresh(self):
        """Refresh main window."""
        logger.log(level=9, msg="MAIN WINDOW: Refreshing")
        self._last_refresh = time.time()
        if self.treemonitor.get_tree() is None:
            logger.log(level=9, msg="No tree to render")
        else:
            logger.log(level=9, msg="Rendering tree")
            self.render_data()
        for win in self.subwindows:
            logger.log(level=9, msg=f"Rendering subwindow: {win}")
            win.render()
        for win in self.subwindows:
            win.refresh()
        self.win.refresh()



"""Provide tests for DelayedSubmission."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import tempfile
from concurrent.futures.process import _ExceptionWithTraceback
from pathlib import Path

import pytest

from joblib_htcondor.delayed_submission import DelayedSubmission


def test_delayed_submission_noargs() -> None:
    """Test DelayedSubmission with no arguments."""

    def myfunc():
        return 100

    with tempfile.TemporaryDirectory() as tmp:
        tmpdir = Path(tmp)
        fname = tmpdir / "test.pickle"
        ds = DelayedSubmission(myfunc)
        assert not ds.done()
        assert ds.args == ()
        assert ds.kwargs == {}
        ds.dump(fname)
        ds.run()
        assert ds.done()
        assert ds.result() == 100
        assert ds.done_timestamp() is not None
        del ds

        ds2 = DelayedSubmission.load(fname)
        assert ds2.args == ()
        assert ds2.kwargs == {}
        assert not ds2.done()
        ds2.run()
        assert ds2.done()
        assert ds2.done_timestamp() is not None
        assert ds2.result() == 100
        fname2 = tmpdir / "test2.pickle"
        ds2.dump(fname2)
        del ds2

        ds3 = DelayedSubmission.load(fname2)
        assert ds3.args == ()
        assert ds3.kwargs == {}
        assert ds3.done()
        assert ds3.result() == 100


def test_delayed_submission_args() -> None:
    """Test DelayedSubmission with arguments."""

    def myfunc(a, b):
        return a + b

    with tempfile.TemporaryDirectory() as tmp:
        tmpdir = Path(tmp)
        fname = tmpdir / "test.pickle"
        ds = DelayedSubmission(myfunc, 10, 20)
        assert not ds.done()
        assert ds.args == (10, 20)
        assert ds.kwargs == {}
        ds.dump(fname)
        ds.run()
        assert ds.done()
        assert ds.done_timestamp() is not None
        assert ds.result() == 30
        del ds

        ds2 = DelayedSubmission.load(fname)
        assert ds2.args == (10, 20)
        assert ds2.kwargs == {}
        assert not ds2.done()
        ds2.run()
        assert ds2.done()
        assert ds2.done_timestamp() is not None
        assert ds2.result() == 30
        fname2 = tmpdir / "test2.pickle"
        ds2.dump(fname2)
        del ds2

        ds3 = DelayedSubmission.load(fname2)
        assert ds3.args == (10, 20)
        assert ds3.kwargs == {}
        assert ds3.done()
        assert ds3.done_timestamp() is not None
        assert ds3.result() == 30


def test_delayed_submission_kwargs() -> None:
    """Test DelayedSubmission with kwarguments."""

    def myfunc(a, b):
        return a + b

    with tempfile.TemporaryDirectory() as tmp:
        tmpdir = Path(tmp)
        fname = tmpdir / "test.pickle"
        ds = DelayedSubmission(myfunc, a=10, b=20)
        assert not ds.done()
        assert ds.args == ()
        assert ds.kwargs == {"a": 10, "b": 20}
        ds.dump(fname)
        ds.run()
        assert ds.done()
        assert ds.done_timestamp() is not None
        assert ds.result() == 30
        del ds

        ds2 = DelayedSubmission.load(fname)
        assert ds2.args == ()
        assert ds2.kwargs == {"a": 10, "b": 20}
        assert not ds2.done()
        assert ds2.done_timestamp() is None
        ds2.run()
        assert ds2.done()
        assert ds2.done_timestamp() is not None
        assert ds2.result() == 30
        fname2 = tmpdir / "test2.pickle"
        ds2.dump(fname2)
        del ds2

        ds3 = DelayedSubmission.load(fname2)
        assert ds3.args == ()
        assert ds3.kwargs == {"a": 10, "b": 20}
        assert ds3.done()
        assert ds3.done_timestamp() is not None
        assert ds3.result() == 30


def test_delayed_submission_allwargs() -> None:
    """Test DelayedSubmission with args and kwargs."""

    def myfunc(a, b):
        return a + b

    with tempfile.TemporaryDirectory() as tmp:
        tmpdir = Path(tmp)
        fname = tmpdir / "test.pickle"
        ds = DelayedSubmission(myfunc, 10, b=20)
        assert not ds.done()
        assert ds.done_timestamp() is None
        assert ds.args == (10,)
        assert ds.kwargs == {"b": 20}
        ds.dump(fname)
        ds.run()
        assert ds.done()
        assert ds.done_timestamp() is not None
        assert ds.result() == 30
        del ds

        ds2 = DelayedSubmission.load(fname)
        assert ds2.args == (10,)
        assert ds2.kwargs == {"b": 20}
        assert not ds2.done()
        ds2.run()
        assert ds2.done()
        assert ds2.done_timestamp() is not None
        assert ds2.result() == 30
        fname2 = tmpdir / "test2.pickle"
        ds2.dump(fname2)
        del ds2

        ds3 = DelayedSubmission.load(fname2)
        assert ds3.args == (10,)
        assert ds3.kwargs == {"b": 20}
        assert ds3.done()
        assert ds3.done_timestamp() is not None
        assert ds3.result() == 30


def test_delayed_submission_error() -> None:
    """Test DelayedSubmission with no arguments."""

    def myfunc():
        raise ValueError("Test error")

    with tempfile.TemporaryDirectory() as tmp:
        tmpdir = Path(tmp)
        fname = tmpdir / "test.pickle"
        ds = DelayedSubmission(myfunc)
        assert not ds.done()
        assert ds.args == ()
        assert ds.kwargs == {}
        ds.run()
        ds.dump(fname)
        assert ds.done()
        assert ds.done_timestamp() is not None
        assert ds.error()
        out = ds.result()

        # Before pickling this should be an _ExceptionWithTraceback
        assert isinstance(out, _ExceptionWithTraceback)
        rebuild, args = out.__reduce__()
        exception = rebuild(*args)  # type: ignore
        assert exception.args == ("Test error",)
        with pytest.raises(ValueError, match="Test error"):
            raise exception
        del ds
        del out
        del exception
        del rebuild
        del args

        # After pickling it should be an exception
        ds2 = DelayedSubmission.load(fname)
        assert ds2.done()
        assert ds2.done_timestamp() is not None
        assert ds2.error()
        out2 = ds2.result()
        assert isinstance(out2, ValueError)
        with pytest.raises(ValueError, match="Test error"):
            raise out2


def test_delayed_submission_results_only() -> None:
    """Test DelayedSubmission with args and kwargs."""

    def myfunc(a, b):
        return a + b

    with tempfile.TemporaryDirectory() as tmp:
        tmpdir = Path(tmp)
        fname = tmpdir / "test.pickle"
        ds = DelayedSubmission(myfunc, 10, b=20)
        assert not ds.done()
        assert ds.done_timestamp() is None
        assert ds.args == (10,)
        assert ds.kwargs == {"b": 20}
        ds.run()
        ds.dump(fname, result_only=True)
        assert ds.done()
        assert ds.done_timestamp() is not None
        assert ds.result() == 30
        assert ds.func is not None
        assert ds.args == (10,)
        assert ds.kwargs == {"b": 20}
        del ds

        ds2 = DelayedSubmission.load(fname)
        assert ds2.done()
        assert ds2.done_timestamp() is not None
        assert ds2.result() == 30
        assert ds2.func is None
        assert ds2.args is None
        assert ds2.kwargs is None



"""Run pickled DelayedSubmission objects."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import time
from datetime import datetime


def logger_level(arg):
    """Parse logging level argument.

    Parameters
    ----------
    arg : str
        The argument to parse.

    Returns
    -------
    int or str
        The parsed argument.

    Raises
    ------
    argparse.ArgumentTypeError
        If the argument cannot be parsed.

    """
    try:
        return int(arg)  # try convert to int
    except ValueError:
        pass
    if arg.upper() in ["DEBUG", "INFO", "WARNING", "ERROR"]:
        return arg.upper()
    raise argparse.ArgumentTypeError(
        "x must be an int or one of 'DEBUG', 'INFO', 'WARNING', 'ERROR'"
    )


if __name__ == "__main__":
    import argparse
    import logging
    import sys
    import warnings
    from pathlib import Path

    from joblib_htcondor.delayed_submission import DelayedSubmission
    from joblib_htcondor.logging import _logging_types, configure_logging

    # Setup logger
    logger = logging.getLogger("joblib_htcondor.executor")
    # Create log stream handler
    lh = logging.StreamHandler(sys.stdout)
    # Create log formatter
    formatter = logging.Formatter(
        "%(asctime)s %(name)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    # Set formatter to handler
    lh.setFormatter(formatter)
    # Add log handler
    logger.addHandler(lh)

    # Create CLI argument parser
    parser = argparse.ArgumentParser(
        description="Run a pickled DelayedSubmission object."
    )
    # Add file argument
    parser.add_argument(
        "filename",
        type=str,
        help="The name of the file to load the DelayedSubmission object from.",
    )
    # Add verbosity argument
    parser.add_argument(
        "--verbose",
        type=logger_level,
        default="INFO",
        help="The logging verbosity to use.",
    )
    # Parse arguments
    args = parser.parse_args()

    # Set logger level
    log_level = args.verbose
    if isinstance(log_level, str):
        log_level = _logging_types[log_level]
    logger.setLevel(log_level)
    logger.info(f"Setting logging level to {args.verbose}")
    configure_logging(level=log_level)

    logger.debug(f"Executor called with {args}")

    # Check and parse file argument
    fname = Path(args.filename)
    if not fname.exists():
        raise FileNotFoundError(f"File {fname} not found.")

    # Create file for run
    run_fname = fname.with_suffix(".run")
    with run_fname.open("w") as f:
        f.write(datetime.now().isoformat())

    # Load file
    logger.info(f"Loading DelayedSubmission object from {fname}")
    ds = None
    while ds is None:
        ds = DelayedSubmission.load(fname)
        if ds is None:
            logger.warning(
                f"Could not load DelayedSubmission object from {fname}. "
                "Retrying in 1 second."
            )
            time.sleep(1)  # Wait 1 second before retrying

    # Issue warning for re-running
    if ds.done():
        warnings.warn(
            "The DelayedSubmission object has already been run.",
            stacklevel=1,
        )
    # Run file
    logger.info("Running DelayedSubmission object")
    ds.run()
    logger.info("DelayedSubmission finished")
    old_stem = fname.stem
    out_fname = fname.with_stem(f"{old_stem}_out")
    logger.info(f"Dumping DelayedSubmission (result only) to {out_fname}")
    # Dump output
    dumped = False
    while not dumped:
        dumped = ds.dump(out_fname, result_only=True)
        if not dumped:
            logger.warning(
                f"Could not dump DelayedSubmission to {out_fname}. "
                "Retrying in 1 second."
            )
            time.sleep(1)
    logger.info("Done.")



"""The joblib htcondor UI Open Menu window."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import curses

from .uilogging import logger
from .utils import (
    align_text,
)
from .window import Window


class OpenMenu(Window):
    """Class for context menu."""

    def __init__(self, parent):
        logger.debug("OPEN MENU: Initializing")
        self.selected = 0
        self.i_first = 0
        self.parent = parent
        self.parent = parent

        self._load_entries()

        y = 1
        w = 57
        x = (self.parent.w - w) // 2
        h = min(len(self.entries) + 2, parent.h - 2)
        logger.debug(f"Creating window {h}x{w} at {y},{x}")
        window = self.parent.win.subwin(h, w, y, x)
        super().__init__(y=y, x=x, window=window)
        self.win.attrset(curses.color_pair(6))

    def _load_entries(self):
        """Load entries."""
        self.selected = 0
        self.i_first = 0
        self.entries = []
        meta_dir = self.parent.get_meta_dir()
        logger.debug(f"Looking for entries in {meta_dir}")
        fnames = list(meta_dir.glob("*-l0.json"))
        for f in fnames:
            logger.debug(f"Adding {f.name}")
            self.entries.append(f.name)

    def resize(self):
        """Handle resize."""
        old_h = self.h
        self.x = (self.parent.w - self.w) // 2
        self.h = min(len(self.entries) + 2, self.parent.h - 2)
        self.win.resize(self.h, self.w)
        self.win.mvderwin(self.y, self.x)
        self.render()
        if old_h > self.h:
            self.parent.render()
        logger.debug(f"OPEN MENU: Resize: {self.h}x57 at {self.y},{self.x}")

    def render(self):
        """Render context menu."""
        logger.debug("OPEN MENU: Rendering")
        self.win.border()
        y, xl, xr = align_text(
            self.win,
            " Open Joblib Tree ",
            0,
            "center",
            curses.color_pair(6),
        )

    def scroll_up(self):
        """Handle scroll up."""
        if self.selected > 0:
            self.selected = self.selected - 1
            logger.debug(f"Scroll up: selected: {self.selected}")
            if self.selected < self.i_first:
                self.i_first -= 1
                logger.debug(f"  Start reached: {self.i_first}")

    def scroll_down(self):
        """Handle scroll down."""
        if self.selected < len(self.entries) - 1:
            self.selected = self.selected + 1
            logger.debug(f"Scroll down: selected: {self.selected}")
            if self.selected - self.i_first >= self.h - 2:
                self.i_first += 1
                logger.debug(f"  End reached: {self.i_first}")

    def refresh(self):
        """Refresh context menu."""
        logger.debug("OPEN MENU: Refreshing")
        max_entries = self.h - 2
        for y, i in enumerate(range(self.i_first, self.i_first + max_entries)):
            if i >= len(self.entries):
                break
            align_text(
                self.win,
                self.entries[i],
                y + 1,
                "center",
                curses.color_pair(15 if i == self.selected else 6),
                fill=-1,
            )
        super().refresh()

    def get_menu(self):
        """Get context menu."""
        return [(0, "Close"), (1, "Clear"), (0, "Refresh")]

    def action(self, action):
        """Trigger action."""
        logger.debug(f"Action: {action}")
        if action == 10:  # Enter
            logger.debug(f"Selected: {self.entries[self.selected]}")
            fpath = self.parent.get_meta_dir() / self.entries[self.selected]
            self.parent.set_path(fpath)
            logger.debug("Self closing menu")
            self.parent.subwindows.pop()
            self.parent.render()
        elif action == ord("c"):
            logger.debug("Closing window")
            self.parent.subwindows.pop()
            self.parent.render()
        elif action == ord("l"):
            self.cleardir()
        elif action == ord("r"):
            self._load_entries()
            self.resize()
            self.render()
        elif action == curses.KEY_UP:
            self.scroll_up()
            self.refresh()
        elif action == curses.KEY_DOWN:
            self.scroll_down()
            self.refresh()

    def cleardir(self):
        """Clear directory."""
        logger.debug("OPEN WINDOW: Clearing directory")
        self.parent.clear_tree()
        meta_dir = self.parent.get_meta_dir()
        data_dir = meta_dir.parent
        for f in meta_dir.glob("*.json"):
            t_data_dir = data_dir / f.stem
            if not t_data_dir.exists():
                logger.debug(f"Removing meta file: {f}")
                f.unlink()
        self._load_entries()
        self.resize()



"""Provide runtime tests."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import logging
import socket

import pytest
from joblib import Parallel, delayed, parallel_config

from joblib_htcondor import register_htcondor


# Check if the test is running on juseless
if socket.gethostname() != "juseless":
    pytest.skip("These tests are only for juseless", allow_module_level=True)


register_htcondor()


def test_normal() -> None:
    """Test normal execution."""
    from operator import neg

    with parallel_config(
        backend="htcondor", n_jobs=-1, request_cpus=1, request_memory="1Gb"
    ):
        out = Parallel()(delayed(neg)(i + 1) for i in range(5))

    assert out == [-1, -2, -3, -4, -5]


def test_exception() -> None:
    """Test running with an exception."""

    def neg_with_exception(a: int) -> int:
        """Negate if a!=2, raise an exception otherwise.

        Parameters
        ----------
        a : int
            The number to negate.

        Returns
        -------
        int
            The negated number.

        Raises
        ------
        ValueError
            If `a` is 2.

        """
        if a == 2:
            raise ValueError("This is an exception")
        else:
            return -a

    with parallel_config(
        backend="htcondor",
        n_jobs=-1,
        request_cpus=1,
        request_memory="1Gb",
        verbose=1000,
        worker_log_level=logging.INFO,
    ):
        with pytest.raises(ValueError, match="This is an exception"):
            Parallel()(delayed(neg_with_exception)(i + 1) for i in range(5))



"""Logging utilities for joblib htcondor."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Sami Hamdan <s.hamdan@fz-juelich.de>
# License: AGPL

import logging
import sys
import warnings
from pathlib import Path
from typing import Optional, Union


__all__ = ["logger"]


logger = logging.getLogger("joblib_htcondor.backend")


_logging_types = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
}


def _close_handlers(logger: logging.Logger) -> None:
    """Safely close relevant handlers for logger.

    Parameters
    ----------
    logger : logging.logger
        The logger to close handlers for.

    """
    for handler in list(logger.handlers):
        if isinstance(handler, (logging.FileHandler, logging.StreamHandler)):
            if isinstance(handler, logging.FileHandler):
                handler.close()
            logger.removeHandler(handler)


def configure_logging(
    level: Union[int, str] = "WARNING",
    fname: Optional[Union[str, Path]] = None,
    overwrite: Optional[bool] = None,
    output_format=None,
) -> None:
    """Configure the logging functionality.

    Parameters
    ----------
    level : int or {"DEBUG", "INFO", "WARNING", "ERROR"}
        The level of the messages to print. If string, it will be interpreted
        as elements of logging (default "WARNING").
    fname : str or pathlib.Path, optional
        Filename of the log to print to. If None, stdout is used
        (default None).
    overwrite : bool, optional
        Overwrite the log file (if it exists). Otherwise, statements
        will be appended to the log (default). None is the same as False,
        but additionally raises a warning to notify the user that log
        entries will be appended (default None).
    output_format : str, optional
        Format of the output messages. See the following for examples:
        https://docs.python.org/dev/howto/logging.html
        e.g., ``"%(asctime)s - %(levelname)s - %(message)s"``.
        If None, default string format is used
        (default ``"%(asctime)s - %(name)s - %(levelname)s - %(message)s"``).

    """
    _close_handlers(logger)  # close relevant logger handlers

    # Set logging level
    if isinstance(level, str):
        level = _logging_types[level]

    # Set logging output handler
    if fname is not None:
        # Convert str to Path
        if not isinstance(fname, Path):
            fname = Path(fname)
        if fname.exists() and overwrite is None:
            warnings.warn(
                f"File ({fname.absolute()!s}) exists. "
                "Messages will be appended. Use overwrite=True to "
                "overwrite or overwrite=False to avoid this message.",
                stacklevel=2,
            )
            overwrite = False
        mode = "w" if overwrite else "a"
        lh = logging.FileHandler(fname, mode=mode)
    else:
        lh = logging.StreamHandler(stream=sys.stdout)  # type: ignore

    # Set logging format
    if output_format is None:
        output_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    # Create logging formatter
    formatter = logging.Formatter(fmt=output_format)

    lh.setFormatter(formatter)  # set formatter
    logger.addHandler(lh)  # set handler
    logger.setLevel(level)  # set level
